"""Translate a pipeline into an Ibis expression."""

import logging
import typing

import ibis

from .ast import ParsedPipeline
from .translation.optimizer import Optimizer
from .translation.options import TranslationOptions
from .translation.steps.add import AddTranslator
from .translation.steps.argmax import ArgMaxTranslator
from .translation.steps.arrayfeatureextractor import ArrayFeatureExtractorTranslator
from .translation.steps.cast import CastLikeTranslator, CastTranslator
from .translation.steps.concat import ConcatTranslator, FeatureVectorizerTranslator
from .translation.steps.div import DivTranslator
from .translation.steps.gather import GatherTranslator
from .translation.steps.identity import IdentityTranslator
from .translation.steps.imputer import ImputerTranslator
from .translation.steps.labelencoder import LabelEncoderTranslator
from .translation.steps.linearclass import LinearClassifierTranslator
from .translation.steps.linearreg import LinearRegressorTranslator
from .translation.steps.matmul import MatMulTranslator
from .translation.steps.mul import MulTranslator
from .translation.steps.onehotencoder import OneHotEncoderTranslator
from .translation.steps.reshape import ReshapeTranslator
from .translation.steps.scaler import ScalerTranslator
from .translation.steps.softmax import SoftmaxTranslator
from .translation.steps.sub import SubTranslator
from .translation.steps.trees import (
    TreeEnsembleClassifierTranslator,
    TreeEnsembleRegressorTranslator,
)
from .translation.steps.where import WhereTranslator
from .translation.steps.zipmap import ZipMapTranslator
from .translation.translator import Translator
from .translation.variables import GraphVariables

# This is a mapping of ONNX operations to their respective translators
# It could be implemented via some form of autodiscovery and
# registration, but explicit mapping avoids effects at a distance and
# makes it easier to understand the translation process.
TRANSLATORS: dict[str, type[Translator]] = {
    "Cast": CastTranslator,
    "CastLike": CastLikeTranslator,
    "Concat": ConcatTranslator,
    "FeatureVectorizer": FeatureVectorizerTranslator,
    "Sub": SubTranslator,
    "MatMul": MatMulTranslator,
    "Add": AddTranslator,
    "Div": DivTranslator,
    "Mul": MulTranslator,
    "Reshape": ReshapeTranslator,
    "Scaler": ScalerTranslator,
    "Gather": GatherTranslator,
    "ArrayFeatureExtractor": ArrayFeatureExtractorTranslator,
    "Identity": IdentityTranslator,
    "Imputer": ImputerTranslator,
    "LabelEncoder": LabelEncoderTranslator,
    "OneHotEncoder": OneHotEncoderTranslator,
    "Where": WhereTranslator,
    "ZipMap": ZipMapTranslator,
    "ArgMax": ArgMaxTranslator,
    "Softmax": SoftmaxTranslator,
    "TreeEnsembleClassifier": TreeEnsembleClassifierTranslator,
    "TreeEnsembleRegressor": TreeEnsembleRegressorTranslator,
    "LinearRegressor": LinearRegressorTranslator,
    "LinearClassifier": LinearClassifierTranslator,
}

log = logging.getLogger(__name__)

# This is primarily for development purposes.
# It's disabled by default because it implies
# a significant cost of executing queries on each step.
LOG_DATA = False
LOG_SQL = False


class ResultsProjection:
    """Projection of the results of the pipeline.

    This class is used to select the columns to be returned
    from the pipeline. It can be used to select specific
    columns to include in the final result set.

    It can also be used to skip the select step of columns
    from the pipeline.

    You can use the `omit` method to skip the projection
    step entirely.
    """

    def __init__(self, select: typing.Optional[list[str]] = None) -> None:
        """
        :param select: A list of additional columns to be selected from the pipeline.
        """
        self._select = select or []
        self._omit = False

    @classmethod
    def omit(cls) -> "ResultsProjection":
        """Create a projection that skips projection phase entirely."""
        projection = cls()
        projection._omit = True
        return projection

    def _expand(self, results: typing.Iterable[str]) -> typing.Optional[list[str]]:
        if self._omit:
            return None

        def _emit_projection() -> typing.Generator[str, None, None]:
            yield from results
            for item in self._select:
                yield item

        return list(_emit_projection())


def translate(
    table: ibis.Table,
    pipeline: ParsedPipeline,
    projection: ResultsProjection = ResultsProjection(),
    *,
    allow_text_tensors: bool = False,
) -> ibis.Table:
    """Translate a pipeline into an Ibis expression.

    Converts the provided ``orbital.ast.ParsedPipeline`` into an Ibis
    expression that reproduces the same transformations. The expression can be
    executed directly or further composed before exporting to SQL.

    :param table: Source ibis table used as the translation input.
    :param pipeline: Parsed pipeline to be translated.
    :param projection: Optional result projection helper.
    :param allow_text_tensors: When ``False`` (default) skip ONNX casts that
        promote numeric/bool tensors to strings solely to coexist with
        passthrough text features. Set to ``True`` to preserve the casts
        exactly as exported.
    """
    optimizer = Optimizer(enabled=True)
    options = TranslationOptions(allow_text_tensors=allow_text_tensors)
    features = {colname: table[colname] for colname in table.columns}
    variables = GraphVariables(features, pipeline._model.graph)
    nodes = list(pipeline._model.graph.node)
    for node in nodes:
        op_type = node.op_type
        if op_type not in TRANSLATORS:
            raise NotImplementedError(f"Translation for {op_type} not implemented")
        translator = TRANSLATORS[op_type](table, node, variables, optimizer, options)  # type: ignore[abstract]
        _log_debug_start(translator, variables)
        translator.process()
        table = translator.mutated_table  # Translator might return a new table.
        _log_debug_end(translator, variables)
    return _projection_results(table, variables, projection)


def _projection_results(
    table: ibis.Table,
    variables: GraphVariables,
    projection: ResultsProjection = ResultsProjection(),
) -> ibis.Table:
    # As we pop out the variables as we use them
    # the remaining ones are the values resulting from all
    # graph branches.
    final_projections: dict[str, typing.Any] = {}
    for key, value in variables.remaining().items():
        if isinstance(value, dict):
            for field in value:
                colkey = key + "." + field
                colvalue = value[field]
                if isinstance(colvalue, ibis.expr.types.StructColumn):
                    raise NotImplementedError(f"StructColumn not supported: {colvalue}")
                final_projections[colkey] = colvalue
        else:
            final_projections[key] = value
    query = table.mutate(**final_projections)
    selection = projection._expand(final_projections.keys())
    if selection is not None:
        query = query.select(*selection)
    return query


def _log_debug_start(translator: Translator, variables: GraphVariables) -> None:
    debug_inputs = {}
    node = translator._node
    for inp in translator._inputs:
        value: typing.Any = None
        if (feature_value := translator._variables.peek_variable(inp)) is not None:
            value = type(feature_value)
        elif (
            initializer := translator._variables.get_initializer_value(inp)
        ) is not None:
            value = initializer
        else:
            raise ValueError(
                f"Unknow input: {inp} for node {node.name}({translator.__class__.__name__})"
            )
        debug_inputs[inp] = value
    log.debug(
        f"Node: {node.name}, Op: {node.op_type}, Attributes: {translator._attributes}, Inputs: {debug_inputs}"
    )
    if LOG_DATA:
        print("Input Data", flush=True)
        print(
            _projection_results(translator.mutated_table, variables).execute(),
            flush=True,
        )
        print("", flush=True)


def _log_debug_end(translator: Translator, variables: GraphVariables) -> None:
    variables = translator._variables
    output_vars = {
        name: type(variables.peek_variable(name)) for name in translator.outputs
    }
    log.debug(
        f"\tOutput: {output_vars} TOTAL: {variables.nested_len()}/{len(variables)}"
    )

    if LOG_DATA:
        print("\tOutput Data", flush=True)
        print(
            _projection_results(translator.mutated_table, variables).execute(),
            flush=True,
        )
        print("", flush=True)
    if LOG_SQL:
        print("\tSQL Expressions", flush=True)
        print(
            ibis.duckdb.connect().compile(
                (_projection_results(translator.mutated_table, variables))
            ),
            flush=True,
        )



"""Export SQL for a pipeline

This module provides a function to export SQL queries for a given pipeline.
It allows to use the prediction pipeline on any supported SQL dialect
without the need for a python runtime environment.
"""

import ibis
import ibis.backends.sql.compilers as sc
import sqlglot.optimizer
import sqlglot.optimizer.optimizer
import sqlglot.schema
from ibis.expr.sql import Catalog

from .ast import ParsedPipeline
from .translate import ResultsProjection, translate

OPTIMIZER_RULES = (
    sqlglot.optimizer.optimizer.qualify,
    sqlglot.optimizer.optimizer.pushdown_projections,
    sqlglot.optimizer.optimizer.normalize,
    sqlglot.optimizer.optimizer.unnest_subqueries,
    sqlglot.optimizer.optimizer.pushdown_predicates,
    sqlglot.optimizer.optimizer.optimize_joins,
    sqlglot.optimizer.optimizer.eliminate_subqueries,
    # sqlglot.optimizer.optimizer.merge_subqueries,  # This makes the SQLGlot optimizer choke with OOMs
    sqlglot.optimizer.optimizer.eliminate_joins,
    sqlglot.optimizer.optimizer.eliminate_ctes,
    sqlglot.optimizer.optimizer.quote_identifiers,
    sqlglot.optimizer.optimizer.canonicalize,
    # sqlglot.optimizer.optimizer.annotate_types,  # This makes the SQLGlot optimizer choke with maximum recursion
    sqlglot.optimizer.optimizer.simplify,
)


def export_sql(
    table_name: str,
    pipeline: ParsedPipeline,
    dialect: str = "duckdb",
    projection: ResultsProjection = ResultsProjection(),
    optimize: bool = True,
    allow_text_tensors: bool = False,
) -> str:
    """Export SQL for a given pipeline.

    Generates a SQL statement equivalent to the provided pipeline for the
    requested dialect. The statement can be executed directly on a database
    that exposes a table matching ``table_name``. Dialect names correspond to
    those listed in ``sqlglot.dialects.DIALECTS``.

    If ``optimize`` is ``True`` the statement is post-processed by sqlglot's
    optimizer, which often produces more compact SQL but may fail on complex
    expressions.

    :param table_name: Name of the source table used in generated SQL.
    :param pipeline: Parsed pipeline to export.
    :param dialect: Target SQL dialect (any supported by sqlglot).
    :param projection: Optional result projection helper.
    :param optimize: Whether to run the sqlglot optimizer (default ``True``).
    :param allow_text_tensors: Forwarded to [orbital.translate.translate][]; controls whether
        numeric/bool tensors coerced to text in ONNX should remain text in the
        resulting SQL. Defaults to ``False`` to keep encoded columns numeric.
    """
    unbound_table = ibis.table(
        schema={
            fname: ftype._to_ibistype() for fname, ftype in pipeline.features.items()
        },
        name=table_name,
    )

    if projection._omit:
        raise ValueError(
            "Projection is empty. Please provide a projection to export SQL."
        )

    ibis_expr = translate(
        unbound_table,
        pipeline,
        projection=projection,
        allow_text_tensors=allow_text_tensors,
    )
    sqlglot_expr = getattr(sc, dialect).compiler.to_sqlglot(ibis_expr)

    if optimize:
        c = Catalog()
        catalog = sqlglot.schema.MappingSchema(
            {unbound_table.get_name(): c.to_sqlglot_schema(unbound_table.schema())},
            normalize=False,
        )
        sqlglot_expr = sqlglot.optimizer.optimize(
            sqlglot_expr, schema=catalog, rules=OPTIMIZER_RULES
        )

    return sqlglot_expr.sql(dialect=dialect)



import ibis
import onnx
import pandas as pd
import pytest
from sklearn.datasets import load_iris
from sklearn.compose import ColumnTransformer
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, TargetEncoder

import orbital
from orbital import types
from orbital.ast import ParsedPipeline

BASIC_FEATURES = {
    "sepal_length": types.FloatColumnType(),
    "sepal_width": types.FloatColumnType(),
    "petal_length": types.FloatColumnType(),
    "petal_width": types.FloatColumnType(),
}
BASIC_MODEL = onnx.helper.make_model(
    onnx.parser.parse_graph("""
agraph (double[?,1] sepal_length, double[?,1] sepal_width, double[?,1] petal_length, double[?,1] petal_width) => (double[?,1] variable) 
   <double[4] Su_Subcst =  {5.84333,3.05733,3.758,1.19933}, double[4,1] coef =  {-0.111906,-0.0400795,0.228645,0.609252}, double[1] intercept =  {1}, int64[2] shape_tensor =  {-1,1}>
{
   merged_columns = Concat <axis: int = 1> (sepal_length, sepal_width, petal_length, petal_width)
   variable1 = Sub (merged_columns, Su_Subcst)
   multiplied = MatMul (variable1, coef)
   resh = Add (multiplied, intercept)
   variable = Reshape (resh, shape_tensor)
}
""")
)


class TestSQLExport:
    @pytest.fixture(scope="class")
    def iris_data(self):
        iris = load_iris()
        # Clean feature names to match what's used in the example
        feature_names = ["sepal_length", "sepal_width", "petal_length", "petal_width"]
        X = pd.DataFrame(iris.data, columns=feature_names)  # Use clean names directly
        y = pd.DataFrame(iris.target, columns=["target"])
        df = pd.concat([X, y], axis=1)
        return df, feature_names

    def test_sql(self):
        parsed_pipeline = ParsedPipeline._from_onnx_model(BASIC_MODEL, BASIC_FEATURES)
        sql = orbital.export_sql("DATA_TABLE", parsed_pipeline, dialect="duckdb")
        assert sql == (
            'SELECT ("t0"."sepal_length" - 5.84333) * -0.111906 + 1.0 + '
            '("t0"."sepal_width" - 3.05733) * -0.0400795 + '
            '("t0"."petal_length" - 3.758) * 0.228645 + '
            '("t0"."petal_width" - 1.19933) * 0.609252 '
            'AS "variable" FROM "DATA_TABLE" AS "t0"'
        )

    def test_sql_optimization_flag(self, iris_data):
        df, feature_names = iris_data

        sklearn_pipeline = Pipeline(
            [("scaler", StandardScaler()), ("regression", LinearRegression())]
        )

        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        optimized_sql = orbital.export_sql(
            "data", parsed_pipeline, dialect="duckdb", optimize=True
        )
        unoptimized_sql = orbital.export_sql(
            "data", parsed_pipeline, dialect="duckdb", optimize=False
        )

        assert (
            optimized_sql
            == 'SELECT 1.0 + ("t0"."sepal_length" - 5.8433332443237305) * -0.1119058608179397432284150575 + ("t0"."sepal_width" - 3.05733323097229) * -0.04007948771815250781921206973 + ("t0"."petal_length" - 3.757999897003174) * 0.2286450295022994613348661968 + ("t0"."petal_width" - 1.1993333101272583) * 0.6092520419738746983614281006 AS "variable.target_0" FROM "data" AS "t0"'
        )
        assert len(optimized_sql) < len(unoptimized_sql)

    @pytest.mark.parametrize(
        "dialect", ["duckdb", "sqlite", "postgres", "mysql", "bigquery", "snowflake"]
    )
    def test_multiple_sql_dialects(self, iris_data, dialect):
        df, feature_names = iris_data

        sklearn_pipeline = Pipeline(
            [("scaler", StandardScaler()), ("regression", LinearRegression())]
        )

        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        try:
            sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)
            assert isinstance(sql, str) and len(sql) > 0
        except Exception as e:
            pytest.skip(f"Dialect {dialect} not supported: {str(e)}")

    def test_sql_columns_passthrough(self, iris_data):
        df, feature_names = iris_data

        sklearn_pipeline = Pipeline(
            [("scaler", StandardScaler()), ("regression", LinearRegression())]
        )

        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        features["ID"] = types.Int64ColumnType(passthrough=True)
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        optimized_sql = orbital.export_sql(
            "data",
            parsed_pipeline,
            dialect="duckdb",
            optimize=True,
            projection=orbital.ResultsProjection(["ID"]),
        )
        assert 'AS "variable.target_0"' in optimized_sql
        assert '"t0"."ID"' in optimized_sql

        unoptimized_sql = orbital.export_sql(
            "data",
            parsed_pipeline,
            dialect="duckdb",
            optimize=False,
            projection=orbital.ResultsProjection(["ID"]),
        )
        assert 'AS "variable.target_0"' in unoptimized_sql
        assert '"t0"."ID"' in unoptimized_sql

    def test_target_encoder_multiple_columns(self):
        df = pd.DataFrame(
            {
                "a": ["a"] * 5 + ["b"] * 5,
                "x": ["a"] * 5 + ["b"] * 5,
            }
        )
        y = [1] * 4 + [0] * 5 + [1]

        pipeline = Pipeline(
            [
                (
                    "encoder",
                    ColumnTransformer(
                        [("te", TargetEncoder(), ["a", "x"])], remainder="passthrough"
                    ),
                )
            ]
        )
        pipeline.fit(df, y)

        features = {name: types.StringColumnType() for name in df.columns}
        parsed_pipeline = orbital.parse_pipeline(pipeline, features=features)

        sql = orbital.export_sql(
            "data", parsed_pipeline, dialect="duckdb", allow_text_tensors=False
        )
        assert '"variable.ordinal_output"' in sql
        assert '"variable.ordinal_output1"' in sql

    def test_target_encoder_outputs_numeric_values(self):
        df = pd.DataFrame(
            {
                "a": ["a"] * 5 + ["b"] * 5,
                "x": ["a"] * 5 + ["b"] * 5,
            }
        )
        y = [1] * 4 + [0] * 5 + [1]

        pipeline = Pipeline(
            [
                (
                    "encoder",
                    ColumnTransformer(
                        [("te", TargetEncoder(), ["a"])], remainder="passthrough"
                    ),
                )
            ]
        )
        pipeline.fit(df, y)

        features = {name: types.StringColumnType() for name in df.columns}
        parsed_pipeline = orbital.parse_pipeline(pipeline, features=features)

        unbound_table = ibis.table(
            {
                fname: ftype._to_ibistype()
                for fname, ftype in parsed_pipeline.features.items()
            },
            name="data",
        )
        translated = orbital.translate(
            unbound_table, parsed_pipeline, allow_text_tensors=False
        )
        schema = translated.schema()
        dtype = schema["transformed_column.variable_cast.ordinal_output"]
        assert dtype.is_numeric()

        sql = orbital.export_sql(
            "data", parsed_pipeline, dialect="duckdb", allow_text_tensors=False
        )
        assert "AS TEXT" not in sql

    def test_allow_text_tensors_toggle(self):
        model = onnx.helper.make_model(
            onnx.helper.make_graph(
                nodes=[
                    onnx.helper.make_node(
                        "Cast",
                        inputs=["input"],
                        outputs=["casted"],
                        name="string_cast",
                        to=onnx.TensorProto.STRING,
                    )
                ],
                name="cast_graph",
                inputs=[
                    onnx.helper.make_tensor_value_info(
                        "input", onnx.TensorProto.FLOAT, [None, 1]
                    )
                ],
                outputs=[
                    onnx.helper.make_tensor_value_info(
                        "casted", onnx.TensorProto.STRING, [None, 1]
                    )
                ],
            )
        )

        features = {"input": types.FloatColumnType()}
        parsed_pipeline = ParsedPipeline._from_onnx_model(model, features)

        sql_numeric = orbital.export_sql(
            "data", parsed_pipeline, dialect="duckdb", allow_text_tensors=False
        )
        assert "CAST(" not in sql_numeric

        sql_text = orbital.export_sql(
            "data", parsed_pipeline, dialect="duckdb", allow_text_tensors=True
        )
        assert "CAST(" in sql_text and "AS TEXT" in sql_text



"""Translators for trees based models."""

from .classifier import TreeEnsembleClassifierTranslator
from .regressor import TreeEnsembleRegressorTranslator

__all__ = ["TreeEnsembleClassifierTranslator", "TreeEnsembleRegressorTranslator"]



import typing

import onnx as _onnx

from .onnx import get_attr_value, get_initializer_data

if typing.TYPE_CHECKING:
    from ..ast import ParsedPipeline


class ParsedPipelineStr:
    def __init__(self, pipeline: "ParsedPipeline", maxlen: int = 80) -> None:
        self._maxlen = maxlen
        self._pipeline = pipeline
        self._constants: dict[str, typing.Any] = {
            init.name: get_initializer_data(init)
            for init in self._pipeline._model.graph.initializer
        }

    def __str__(self) -> str:
        """Generate a string representation of the pipeline."""
        return f"""{self._pipeline.__class__.__name__}(
  features={{\n{self._features_str()}\n  }},
  steps=[\n{self._steps_str()}\n  ],
)
"""

    def _features_str(self) -> str:
        """Generate a string representation of the features."""
        return "\n".join(
            (
                f"    {feature_name}: {feature_type}"
                for feature_name, feature_type in self._pipeline.features.items()
            )
        )

    def _steps_str(self) -> str:
        """Generate a string representation of the pipeline steps."""
        return "\n".join(
            (self._node_str(node) for node in self._pipeline._model.graph.node)
        )

    def _node_str(self, node: _onnx.NodeProto) -> str:
        """Generate a string representation of a pipeline step."""
        return f"""    {self._varnames(node.output)}={node.op_type}(
      inputs: {self._varnames(node.input)},
      attributes: {self._attributes(node.attribute)}
    )"""

    def _varnames(self, varlist: typing.Iterable[str]) -> str:
        """Generate a string representation of a list of variables or constants."""

        def _var_value(var: str) -> str:
            if var in self._constants:
                return self._shorten(f"{var}={self._constants[var]}")
            return f"{var}"

        return ", ".join((f"{_var_value(var)}" for var in varlist))

    def _attributes(self, attributes: typing.Iterable[_onnx.AttributeProto]) -> str:
        """Generate a string representation of a list of attributes."""

        def _attr_value(attr: _onnx.AttributeProto) -> str:
            return self._shorten(str(get_attr_value(attr)))

        indent = "\n        "
        content = indent.join(
            (f"{attr.name}={_attr_value(attr)}" for attr in attributes)
        )
        if content.strip():
            return f"{indent}{content}"
        else:
            return ""

    def _shorten(self, value: str) -> str:
        """Shorten a string to maxlen characters."""
        if self._maxlen and len(value) > self._maxlen:
            return f"{value[: self._maxlen]}..."
        return value



"""Post-transformation functions for ONNX operators.

This module provides post-transformation functions that can be applied to
both individual numeric values and groups of numeric variables.

Example usage:
    # For a single numeric value
    score = ibis.literal(2.0)
    transformed_score = apply_post_transform(score, "LOGISTIC")
    # Result: Divide(1, Add(Exp(Negate(2.0)), 1)): 1 / Exp(Negate(2.0)) + 1

    # For a group of variables
    scores_group = NumericVariablesGroup({
        "class_0": ibis.literal(1.0),
        "class_1": ibis.literal(-0.5)
    })
    transformed_group = apply_post_transform(scores_group, "LOGISTIC")
    # Result: NumericVariablesGroup({
    #     "class_0": Divide(1, Add(Exp(Negate(1.0)), 1)),
    #     "class_1": Divide(1, Add(Exp(Negate(-0.5)), 1))
    # })
    # Mathematical values: class_0: 1.0 → 0.731059, class_1: -0.5 → 0.377541

    # SOFTMAX transformation (only for groups)
    softmax_group = apply_post_transform(scores_group, "SOFTMAX")
    # Result: NumericVariablesGroup with normalized probabilities that sum to 1.0
"""

import abc
import typing

import ibis

from .variables import NumericVariablesGroup, ValueVariablesGroup


def apply_post_transform(
    data: typing.Union[
        ibis.expr.types.NumericValue, NumericVariablesGroup, ValueVariablesGroup
    ],
    transform: str,
) -> typing.Union[ibis.expr.types.NumericValue, NumericVariablesGroup]:
    """Apply post-transformation to a single numeric value or a group of numeric variables.

    :param data: Either a single NumericValue, NumericVariablesGroup, or ValueVariablesGroup containing NumericValues
    :param transform: Name of the transformation to apply (e.g., "LOGISTIC", "NONE", "SOFTMAX")
    :return: The transformed data in the same format as the input
    """
    transform_class = TRANSFORM_CLASSES.get(transform)
    if not transform_class:
        raise NotImplementedError(f"Post transform '{transform}' is not implemented.")

    if isinstance(data, ValueVariablesGroup):
        # Convert ValueVariablesGroup to NumericVariablesGroup
        # this is mostly for convenience, it will raise TypeError if any value is not numeric
        data = NumericVariablesGroup(data)

    if transform_class is IdentityTransform:
        # Identity transform does not change the data
        # We can avoid applying it at all.
        return data

    transformer = transform_class()

    if isinstance(data, NumericVariablesGroup):
        # Apply transformation to the group of variables
        return transformer.transform_variables_group(data)
    elif isinstance(data, ibis.expr.types.NumericValue):
        # Apply transformation to single value
        return transformer.transform_numeric(data)
    else:
        raise TypeError(
            f"Expected NumericValue or NumericVariablesGroup, got {type(data)}"
        )


class PostTransform(abc.ABC):
    """Base class for post-transformation operations."""

    @abc.abstractmethod
    def transform_numeric(
        self, value: ibis.expr.types.NumericValue
    ) -> ibis.expr.types.NumericValue:
        """Transform a single numeric value.

        :param value: The numeric value to transform
        :return: The transformed numeric value
        """
        pass

    def transform_variables_group(
        self, group: NumericVariablesGroup
    ) -> NumericVariablesGroup:
        """Transform a group of numeric variables.

        Default implementation applies transform_numeric to each variable in the group.
        Subclasses can override this for transformations that need to consider all variables together.

        :param group: The group of numeric variables to transform
        :return: The transformed group of numeric variables
        """
        return NumericVariablesGroup(
            {name: self.transform_numeric(value) for name, value in group.items()}
        )


class LogisticTransform(PostTransform):
    """Logistic (sigmoid) transformation: 1 / (1 + exp(-x))."""

    def transform_numeric(
        self, value: ibis.expr.types.NumericValue
    ) -> ibis.expr.types.NumericValue:
        """Apply logistic transformation to a single value."""
        return 1 / (1 + (-value).exp())


class SoftmaxTransform(PostTransform):
    """Softmax transformation: exp(x_i) / sum(exp(x_j)) for all j."""

    def transform_numeric(
        self, value: ibis.expr.types.NumericValue
    ) -> ibis.expr.types.NumericValue:
        """Apply softmax to a single value: exp(x) / exp(x) = 1.0."""
        return ibis.literal(1.0)

    def transform_variables_group(
        self, group: NumericVariablesGroup
    ) -> NumericVariablesGroup:
        """Apply softmax transformation across all variables in the group.

        Uses numerical stability: subtract max before exponentiation.
        """
        # Use numerical stability: subtract max before exponentiation
        max_value = ibis.greatest(*group.values())

        # Compute exp(x_i - max) for each variable
        exp_dict = {name: (value - max_value).exp() for name, value in group.items()}

        # Sum all exponentials
        sum_exp = sum(exp_dict.values())

        # Compute softmax: exp(x_i - max) / sum(exp(x_j - max))
        return NumericVariablesGroup(
            {name: exp_val / sum_exp for name, exp_val in exp_dict.items()}
        )


class NormalizeTransform(PostTransform):
    """Normalization transformation: x / TOTAL where TOTAL is the sum of all variables."""

    def transform_numeric(
        self, value: ibis.expr.types.NumericValue
    ) -> ibis.expr.types.NumericValue:
        """Apply normalization to a single value."""
        return 1.0

    def transform_variables_group(
        self, group: NumericVariablesGroup
    ) -> NumericVariablesGroup:
        """Apply L1 normalization across all variables in the group."""
        sum_votes = sum(group.values())
        return NumericVariablesGroup(
            {name: value / sum_votes for name, value in group.items()}
        )


class IdentityTransform(PostTransform):
    """Identity transformation (no change).

    This primarily only acts as a guard
    against applying unnecessary transformations.
    It still provides a working implementation,
    but `apply_post_transform` will ignore it.
    """

    def transform_numeric(
        self, value: ibis.expr.types.NumericValue
    ) -> ibis.expr.types.NumericValue:
        """Return the value unchanged."""
        return value


# Mapping of transformation names to their corresponding classes
TRANSFORM_CLASSES: dict[str, type[PostTransform]] = {
    "LOGISTIC": LogisticTransform,
    "NONE": IdentityTransform,
    "SOFTMAX": SoftmaxTransform,
    # Make sure you prefix ORBITAL specific transforms with ORBITAL_
    # to avoid conflicts with ONNX transforms.
    "ORBITAL_NORMALIZE": NormalizeTransform,
}



"""Implementation of the LinearRegression operator."""

import typing

import ibis

from ..translator import Translator
from ..variables import NumericVariablesGroup, ValueVariablesGroup, VariablesGroup


class LinearRegressorTranslator(Translator):
    """Processes a LinearRegression node and updates variables with the predicted expression.

    The LinearRegression operator computes predictions as:
    Y = X * coefficients + intercept

    For more complex pipelines the LinearRegression operator is not always used,
    usually a combination of Mul and Add operations is used.
    """

    def process(self) -> None:
        """Performs the translation and sets the output variable."""
        # https://onnx.ai/onnx/operators/onnx_aionnxml_LinearRegressor.html
        coefficients = typing.cast(list[float], self._attributes["coefficients"])
        intercepts = typing.cast(list[float], self._attributes.get("intercepts", [0.0]))
        targets = typing.cast(int, self._attributes.get("targets", 1))

        post_transform = self._attributes.get("post_transform", "NONE")
        if post_transform != "NONE":
            raise NotImplementedError("Post transform is not implemented.")

        if len(intercepts) not in [0, targets]:
            raise ValueError(
                "LinearRegressor: intercepts length must match targets or be empty."
            )

        if len(self._inputs) != 1:
            raise ValueError("LinearRegressor node must have exactly 1 input.")

        input_operand = self._variables.consume(self._inputs[0])

        if isinstance(input_operand, VariablesGroup):
            input_operand = NumericVariablesGroup(input_operand)
            num_features = len(input_operand)

            if len(coefficients) != targets * num_features:
                raise ValueError(
                    "Coefficients length must equal targets number of input fields."
                )

            results = {}
            fields = list(input_operand.values())

            for target_idx in range(targets):
                start = target_idx * num_features
                end = start + num_features
                coef_slice = coefficients[start:end]

                intercept = intercepts[target_idx] if intercepts else 0.0

                prediction = ibis.literal(intercept)
                for val, coef in zip(fields, coef_slice):
                    prediction += val * coef

                # TODO: apply post_transform here if needed

                results[f"target_{target_idx}"] = self._optimizer.fold_operation(
                    prediction
                )

            self.set_output(ValueVariablesGroup(results))

        else:
            input_operand = typing.cast(ibis.expr.types.NumericValue, input_operand)

            if targets != 1 or len(coefficients) != 1:
                raise ValueError(
                    "Single column input expects exactly one target and one coefficient."
                )

            intercept = intercepts[0] if intercepts else 0.0
            prediction = (input_operand * coefficients[0]) + intercept

            # TODO: apply post_transform here if needed

            self.set_output(self._optimizer.fold_operation(prediction))



"""Implement optiomizations to the Ibis expression tree.

Primarily it takes care of folding constant expressions
and removing unnecessary casts.
"""

import itertools
import operator
import typing

import ibis
import ibis.expr.datatypes as dt
from ibis.expr.operations import (
    Abs,
    Add,
    And,
    Binary,
    Ceil,
    Divide,
    Equals,
    Floor,
    FloorDivide,
    Greater,
    GreaterEqual,
    IdenticalTo,
    Less,
    LessEqual,
    Literal,
    Modulus,
    Multiply,
    Negate,
    Not,
    NotEquals,
    Or,
    Subtract,
    Unary,
    Xor,
)
from ibis.expr.types import NumericScalar

# Alias to avoid long references when checking for nested casts.
CastOp = ibis.expr.operations.Cast


class Optimizer:
    """Optimizer for Ibis expressions.

    This class is responsible for applying a set of optimization
    processes to Ibis expressionsto remove unecessary operations and
    reduce query complexity.
    """

    BINARY_OPS: dict[type[Binary], typing.Callable] = {
        # Mathematical Operators
        Add: operator.add,
        Subtract: operator.sub,
        Multiply: operator.mul,
        Divide: operator.truediv,
        FloorDivide: operator.floordiv,
        Modulus: operator.mod,
        # Logical Operators
        Equals: operator.eq,
        NotEquals: operator.ne,
        Greater: operator.gt,
        GreaterEqual: operator.ge,
        Less: operator.lt,
        LessEqual: operator.le,
        IdenticalTo: operator.eq,
        # Binary Operators
        And: operator.and_,
        Or: operator.or_,
        Xor: operator.xor,
    }

    UNARY_OPS: dict[type[Unary], typing.Callable] = {
        Negate: operator.neg,
        Abs: operator.abs,
        Ceil: lambda x: float(operator.methodcaller("ceil")(x)),  # Se necessario
        Floor: lambda x: float(operator.methodcaller("floor")(x)),
        Not: operator.not_,
    }

    def __init__(self, enabled: bool = True) -> None:
        """
        :param enabled: Whether to enable the optimizer.
                        When disabled, the optimizer will
                        return the expression unchanged.
        """
        self.ENABLED = enabled

    def _ensure_expr(self, value: ibis.Expr) -> ibis.Expr:
        """Ensure that the value is an Ibis expression.

        Literal objects need to be converted back to an
        Ibis expression to be used in the query.
        """
        if isinstance(value, Literal):
            return ibis.literal(value.value)
        return value

    def _fold_associative_op_contiguous(
        self, lst: list[ibis.expr.types.NumericValue], pyop: typing.Callable
    ) -> list[ibis.expr.types.NumericValue]:
        """Precompute an operation applied on multiple elements.

        Given a list of expressions and a binary operation,
        this function will precompute the operation on all
        constant expressions in the list returning a new list
        of expressions with the folded constants.
        """
        if self.ENABLED is False:
            return list(lst)

        if pyop not in {operator.add, operator.mul}:
            raise NotImplementedError(
                "Only addition and multiplication folding are supported."
            )

        resulting_items = []
        items = list(lst)
        folded_value = None

        while items:
            expr = items.pop(0)
            if isinstance(expr, NumericScalar):
                value = expr.op().value
                if folded_value is None:
                    folded_value = value
                else:
                    folded_value = pyop(folded_value, value)
            else:
                resulting_items.append(expr)

        if folded_value is not None:
            resulting_items.append(ibis.literal(folded_value))

        return resulting_items

    def fold_contiguous_sum(
        self, lst: list[ibis.expr.types.NumericValue]
    ) -> list[ibis.expr.types.NumericValue]:
        """Precompute constants in a list of sums"""
        return self._fold_associative_op_contiguous(lst, operator.add)

    def fold_contiguous_product(
        self, lst: list[ibis.expr.types.NumericValue]
    ) -> list[ibis.expr.types.NumericValue]:
        """Precompute constants in a list of multiplications"""
        return self._fold_associative_op_contiguous(lst, operator.mul)

    def fold_case(self, expr: typing.Union[ibis.Value, ibis.Deferred]) -> ibis.Value:
        """Apply different folding strategies to CASE WHHEN expressions.

        - If the CASE is a constant, it will evalute it immediately.
        - If the CASE is a IF ELSE statement returning 1 or 0,
          it will be converted to a boolean expression.
        - When the results and the default are the same, just return
          the default value.
        """
        if not isinstance(expr, ibis.Value):
            raise NotImplementedError("Deferred case expressions are not supported")

        if self.ENABLED is False:
            return expr

        op = expr.op()

        results_are_literals = all(
            isinstance(c, Literal) for c in itertools.chain([op.default], op.results)
        )
        possible_values = (
            set(itertools.chain([op.default.value], [c.value for c in op.results]))
            if results_are_literals
            else set()
        )

        if results_are_literals and len(possible_values) == 1:
            # All results and the default are literals with the same value.
            # It doesn't make any sense to have the case as it will always
            # lead to the same result.
            return self._ensure_expr(possible_values.pop())
        elif len(op.cases) == 1 and isinstance(op.cases[0], Literal):
            # It's only a IF ELSE statement, we can check the case
            # and eventually drop it if it's a constant.
            if op.cases[0].value:
                return op.results[0].to_expr()
            else:
                return op.default.to_expr()
        elif len(op.cases) == 1 and results_are_literals and possible_values == {1, 0}:
            # results are 1 or 0, we can fold it to a boolean expression.
            # FIXME: This doesn't work on postgresql so we need to disable it for the moment.
            return expr
            if op.results[0].value == 1:
                return (op.cases[0].to_expr()).cast("float64")
            else:
                return (~(op.cases[0].to_expr())).cast("float64")

        return expr

    def fold_cast(self, expr: ibis.Value) -> ibis.Value:
        """Given a cast expression, precompute it if possible."""
        if self.ENABLED is False:
            return expr

        op_instance = expr.op()
        if not isinstance(op_instance, CastOp):
            # Not a cast, ignore
            # This can happen when a Field (a column) is casted to a type
            # and the Column is already of the same type.
            # Ibis seems to optimize this case and remove the cast.
            return expr

        target_type = op_instance.to
        arg_op = op_instance.arg

        # Collapse nested casts so we only apply the outermost cast once.
        while isinstance(arg_op, CastOp):
            arg_op = arg_op.arg

        if isinstance(arg_op, Literal):
            value = arg_op.value
            if target_type == dt.int64:
                return ibis.literal(int(value))
            elif target_type == dt.float64:
                return ibis.literal(float(value))
            elif target_type == dt.string:
                return ibis.literal(str(value))
            elif target_type == dt.boolean:
                return ibis.literal(bool(value))
            else:
                raise NotImplementedError(
                    f"Literal Cast to {target_type} not supported"
                )

        arg_expr = arg_op.to_expr()
        if arg_expr.type() == target_type:
            # The expression is already of the target type
            # No need to cast it again.
            return arg_expr

        return arg_expr.cast(target_type)

    def fold_zeros(self, expr: ibis.Expr) -> ibis.Expr:
        """Given a binary expression, precompute the result if it contains zeros.

        Operations like x + 0, x * 0, x - 0 etc can be folded in just x or 0
        without the need to compute the operation.
        """
        if self.ENABLED is False:
            return expr

        op = expr.op()
        inputs = op.args
        op_class = type(op)

        if op_class == Multiply:
            left_val = inputs[0].value if isinstance(inputs[0], Literal) else None
            right_val = inputs[1].value if isinstance(inputs[1], Literal) else None
            if left_val == 0 or right_val == 0:
                return ibis.literal(0)
        elif op_class == Add:
            left_val = inputs[0].value if isinstance(inputs[0], Literal) else None
            right_val = inputs[1].value if isinstance(inputs[1], Literal) else None
            if left_val == 0:
                return inputs[1].to_expr()
            elif right_val == 0:
                return inputs[0].to_expr()
        elif op_class == Subtract:
            left_val = inputs[0].value if isinstance(inputs[0], Literal) else None
            right_val = inputs[1].value if isinstance(inputs[1], Literal) else None
            if left_val == 0:
                return inputs[1].to_expr()
            elif right_val == 0:
                return inputs[0].to_expr()

        return expr

    def fold_operation(self, expr: ibis.Expr) -> ibis.Expr:
        """Given a node (an Ibis expression) fold constant expressions.

        If all node immediate children are constant (i.e. NumericScalar),
        compute the operation in Python and return a literal with the result.

        Otherwise, simply return the expression unchanged.

        This function assumes that constant folding has already been applied
        to the children.
        """
        if self.ENABLED is False:
            return expr

        if isinstance(expr, (int, float, str, bool)):
            # In some cases the operation has been computed in python.
            # For example when we try to compute * between a ONNX literal
            # and a previously folded expression.
            # In those case return a literal so we guarantee we always
            # return an Ibis expression
            return ibis.literal(expr)

        op = expr.op()
        inputs = op.args

        if not all(isinstance(child, Literal) for child in inputs):
            # We can only fold operations where all children are literals.
            # At least we can remove zeros if they exist.
            return self.fold_zeros(expr)

        op_class = type(op)
        if op_class in self.BINARY_OPS:
            left_val = inputs[0].value
            right_val = inputs[1].value
            result = self.BINARY_OPS[typing.cast(type[Binary], op_class)](
                left_val, right_val
            )
            return self._ensure_expr(result)
        elif op_class in self.UNARY_OPS and len(inputs) == 1:
            result = self.UNARY_OPS[typing.cast(type[Unary], op_class)](inputs[0].value)
            return self._ensure_expr(result)
        else:
            # No possible folding
            return expr

    def _debug(self, expr: ibis.Expr, show_args: bool = True) -> str:
        """Given an expression, return a string representation for debugging."""
        if isinstance(expr, Literal):
            return repr(expr.value)
        elif show_args is False:
            return type(expr).__name__
        elif not hasattr(expr, "args"):
            return f"{type(expr).__name__}(<unknown>)"
        else:
            return f"{type(expr).__name__}({', '.join([self._debug(a, show_args=False) for a in expr.args])})"



from unittest import mock

import pandas as pd
import polars as pl
import pyarrow as pa
import pytest

from orbital import types


class TestDataTypesGuessing:
    DF_DATA = {
        "names": ["Aldo", "Giovanni", "Giacomo"],
        "age": [66, 68, 68],
        "ratio": [66 / 66, 68 / 66, 68 / 66],
    }
    EXPECTED_TYPES = {
        "names": types.StringColumnType(),
        "age": types.Int64ColumnType(),
        "ratio": types.DoubleColumnType(),
    }

    def test_from_pandas(self):
        df = pd.DataFrame(self.DF_DATA)
        assert types.guess_datatypes(df) == self.EXPECTED_TYPES

    def test_from_polars(self):
        df = pl.DataFrame(self.DF_DATA)
        assert types.guess_datatypes(df) == self.EXPECTED_TYPES

    def test_from_pyarrow(self):
        df = pa.table(self.DF_DATA)
        assert types.guess_datatypes(df) == self.EXPECTED_TYPES

    def test_invalid_datatype(self):
        with pytest.raises(ValueError) as exc:
            types.guess_datatypes({"column": 5})
        assert exc.match("Unable to guess types of dataframe")

    def test_invalid_datatype_conversion(self):
        with mock.patch.object(
            types._sl2o_types, "guess_data_type", return_value=[("somecol", "invalid")]
        ):
            with pytest.raises(ValueError) as exc:
                types.guess_datatypes("Doesn't matter")
            assert exc.match("Unsupported datatype for column somecol")

    def test_alltypes(self):
        for t in [
            types.FloatColumnType,
            types.Float16ColumnType,
            types.DoubleColumnType,
            types.StringColumnType,
            types.Int64ColumnType,
            types.UInt64ColumnType,
            types.Int32ColumnType,
            types.UInt32ColumnType,
            types.Int16ColumnType,
            types.UInt16ColumnType,
            types.Int8ColumnType,
            types.UInt8ColumnType,
            types.BooleanColumnType,
        ]:
            onxtype = t()._to_onnxtype()
            assert isinstance(onxtype, types._sl2o_types.DataType)
            assert t._from_onnxtype(onxtype) == t()

    def test_only_support_column_types(self):
        with pytest.raises(ValueError) as exc:
            types.ColumnType._from_onnxtype(
                types._sl2o_types.FloatTensorType(shape=[1, 1])
            )
        assert exc.match("Only columnar data is supported")

    def test_invalid_datatype_shape(self):
        with pytest.raises(TypeError) as exc:
            types.ColumnType._from_onnxtype(mock.Mock(shape=[None, 1]))
        assert exc.match("Unsupported data type Mock")



"""Implementation of the Scaler operator."""

import typing

import ibis

from ..translator import Translator
from ..variables import NumericVariablesGroup, ValueVariablesGroup, VariablesGroup


class ScalerTranslator(Translator):
    """Processes a Scaler node and updates variables with the scaled expression.

    The Scaler operator applies a scaling and offset to the input:
    Y = (X - offset) * scale

    The scaler operation is not always used, for more complex pipelines
    usually a combination of Sub and Mul operations is used.
    """

    def process(self) -> None:
        """Performs the translation and sets the output variable."""
        # https://onnx.ai/onnx/operators/onnx_aionnxml_Scaler.html
        scale = typing.cast(list[float], self._attributes["scale"])
        offset = typing.cast(list[float], self._attributes["offset"])

        if len(self._inputs) != 1:
            raise ValueError("Scaler node must have exactly 1 input.")

        input_operand = self._variables.consume(self._inputs[0])

        type_check_var = input_operand
        if isinstance(type_check_var, dict):
            type_check_var = next(iter(type_check_var.values()), None)

        if not isinstance(type_check_var, ibis.expr.types.NumericValue):
            raise ValueError("Scaler: The input operand must be numeric.")

        if isinstance(input_operand, VariablesGroup):
            input_operand = NumericVariablesGroup(input_operand)

            # If the attributes are len=1,
            # it means to apply the same value to all inputs.
            num_fields = len(input_operand)
            if len(offset) == 1:
                offset = offset * num_fields
            if len(scale) == 1:
                scale = scale * num_fields

            if len(offset) != num_fields or len(scale) != num_fields:
                raise ValueError(
                    "Scaler: offset and scale lists must match the number of input fields."
                )

            self.set_output(
                ValueVariablesGroup(
                    {
                        field: self._optimizer.fold_operation(
                            (val - offset[i]) * scale[i]
                        )
                        for i, (field, val) in enumerate(input_operand.items())
                    }
                )
            )
        else:
            input_operand = typing.cast(ibis.expr.types.NumericValue, input_operand)
            self.set_output(
                self._optimizer.fold_operation((input_operand - offset[0]) * scale[0])
            )



import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.feature_selection import SelectKBest, f_regression
from sklearn.linear_model import ElasticNet, LinearRegression, LogisticRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder, StandardScaler

import orbital
from orbital import types
from orbital_testing_helpers import execute_sql


class TestEndToEndPipelines:
    def test_simple_linear_regression(self, iris_data, db_connection):
        df, feature_names = iris_data
        conn, dialect = db_connection

        sklearn_pipeline = Pipeline(
            [("scaler", StandardScaler()), ("regression", LinearRegression())]
        )
        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_preds = sklearn_pipeline.predict(X)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)

        sql_results = execute_sql(sql, conn, dialect, df)
        np.testing.assert_allclose(
            sql_results.values.flatten(), sklearn_preds.flatten(), rtol=1e-4, atol=1e-4
        )

    def test_simple_linear_with_projection(self, iris_data, db_connection):
        df, feature_names = iris_data
        conn, dialect = db_connection

        sklearn_pipeline = Pipeline(
            [("scaler", StandardScaler()), ("regression", LinearRegression())]
        )
        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_preds = sklearn_pipeline.predict(X)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql(
            "data",
            parsed_pipeline,
            projection=orbital.ResultsProjection(["sepal_length"]),
            dialect=dialect,
        )

        sql_results = execute_sql(sql, conn, dialect, df)
        assert set(sql_results.columns) == {"sepal_length", "variable.target_0"}
        np.testing.assert_allclose(
            sql_results["variable.target_0"].values.flatten(),
            sklearn_preds.flatten(),
            rtol=1e-4,
            atol=1e-4,
        )

    def test_feature_selection_pipeline(self, diabetes_data, db_connection):
        df, feature_names = diabetes_data
        conn, dialect = db_connection

        sklearn_pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),
                ("feature_selection", SelectKBest(f_regression, k=5)),
                ("regression", LinearRegression()),
            ]
        )
        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_preds = sklearn_pipeline.predict(X)

        features = {str(fname): types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)

        sql_results = execute_sql(sql, conn, dialect, df)
        np.testing.assert_allclose(
            sql_results.values.flatten(), sklearn_preds.flatten(), rtol=1e-4, atol=1e-4
        )

    def test_column_transformer_pipeline(self, iris_data, db_connection):
        df, feature_names = iris_data
        conn, dialect = db_connection

        df["cat_feature"] = np.random.choice(["A", "B", "C"], size=df.shape[0])

        preprocessor = ColumnTransformer(
            transformers=[
                ("num", StandardScaler(), feature_names),
                ("cat", OneHotEncoder(), ["cat_feature"]),
            ]
        )

        sklearn_pipeline = Pipeline(
            [("preprocessor", preprocessor), ("regression", LinearRegression())]
        )

        X = df[feature_names + ["cat_feature"]]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_preds = sklearn_pipeline.predict(X)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        features["cat_feature"] = types.StringColumnType()
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)

        sql_results = execute_sql(sql, conn, dialect, df)
        np.testing.assert_allclose(
            sql_results.values.flatten(), sklearn_preds.flatten(), rtol=1e-4, atol=1e-4
        )

    def test_logistic_regression(self, iris_data, db_connection):
        df, feature_names = iris_data
        conn, dialect = db_connection

        binary_df = df[df["target"].isin([0, 1])].copy()

        sklearn_pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),
                ("classifier", LogisticRegression(random_state=42)),
            ]
        )

        X = binary_df[feature_names]
        y = binary_df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_proba = sklearn_pipeline.predict_proba(X)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)

        sql_results = execute_sql(sql, conn, dialect, binary_df)

        sklearn_proba_df = pd.DataFrame(
            sklearn_proba, columns=sklearn_pipeline.classes_, index=binary_df.index
        )

        for class_label in sklearn_pipeline.classes_:
            np.testing.assert_allclose(
                sql_results[f"output_probability.{class_label}"].values.flatten(),
                sklearn_proba_df[class_label].values.flatten(),
                rtol=1e-4,
                atol=1e-4,
            )

    def test_elasticnet(self, diabetes_data, db_connection):
        """Test an ElasticNet pipeline with preprocessing transformations."""
        df, feature_names = diabetes_data
        conn, dialect = db_connection

        sklearn_pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),  # Standardize features
                ("normalizer", MinMaxScaler()),  # Scale to [0,1] range
                ("regressor", ElasticNet(alpha=0.5, l1_ratio=0.5, random_state=42)),
            ]
        )

        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_preds = sklearn_pipeline.predict(X)

        features = {str(fname): types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)
        sql_results = execute_sql(sql, conn, dialect, df)

        np.testing.assert_allclose(
            sql_results["variable.target_0"].to_numpy(),
            sklearn_preds.flatten(),
            rtol=1e-4,
            atol=1e-4,
        )



"""Implementation of the Softmax operator."""

import typing

import ibis

from ..transformations import apply_post_transform
from ..translator import Translator
from ..variables import NumericVariablesGroup, VariablesGroup


class SoftmaxTranslator(Translator):
    """Processes a Softmax node and updates the variables with the output expression.

    The operation computes the normalized exponential of the input::

        Softmax = Exp(input) / Sum(Exp(input))

    Currently the Softmax operation is supported only for axis=-1 or axis=1,
    which means for the a column group means that the softmax is computed
    independently for each column in the group.
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx__Softmax.html
        data = self._variables.consume(self.inputs[0])
        if not isinstance(data, (ibis.expr.types.NumericValue, dict)):
            raise ValueError(
                "Softmax: The first operand must be a numeric column or a column group of numerics."
            )

        axis = self._attributes.get("axis", -1)
        if axis not in (-1, 1):
            raise ValueError(
                "SoftmaxTranslator supports only axis=-1 or axis=1 for group of columns"
            )

        if isinstance(data, VariablesGroup):
            data = NumericVariablesGroup(data)
        else:
            data = typing.cast(ibis.expr.types.NumericValue, data)
        self.set_output(apply_post_transform(data, "SOFTMAX"))



import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.datasets import make_classification
from sklearn.ensemble import (
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
)
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder, StandardScaler
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

import orbital
from orbital import types
from orbital_testing_helpers import execute_sql


class TestTreeBasedPipelines:
    """Test suite for tree-based machine learning pipelines and their SQL exports."""

    # Decision Tree Tests
    def test_decision_tree_classifier(self, iris_data, db_connection):
        """Test a decision tree classifier pipeline with preprocessing."""
        df, _ = iris_data
        conn, dialect = db_connection

        # Use binary classification for simplicity
        binary_df = df[df["target"].isin([0, 1])].copy()
        binary_df = pd.concat([binary_df.iloc[:10], binary_df.iloc[-10:]])
        binary_df = binary_df.reset_index(drop=True)

        # Add StandardScaler as preprocessing step
        sklearn_pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),  # Normalize features
                ("classifier", DecisionTreeClassifier(max_depth=3, random_state=42)),
            ]
        )

        X = binary_df["petal_length"].to_frame()
        y = binary_df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_proba = sklearn_pipeline.predict_proba(X)
        sklearn_class = sklearn_pipeline.predict(X)

        features = {fname: types.FloatColumnType() for fname in ["petal_length"]}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)
        sql_results = execute_sql(sql, conn, dialect, binary_df)

        sklearn_proba_df = pd.DataFrame(
            sklearn_proba, columns=sklearn_pipeline.classes_, index=binary_df.index
        )

        np.testing.assert_allclose(
            sql_results["output_label"].to_numpy(), sklearn_class
        )
        for class_label in sklearn_pipeline.classes_:
            np.testing.assert_allclose(
                sql_results[f"output_probability.{class_label}"].values.flatten(),
                sklearn_proba_df[class_label].values.flatten(),
            )

    def test_decision_tree_regressor(self, iris_data, db_connection):
        """Test a decision tree regressor pipeline with feature selection."""
        df, feature_names = iris_data
        conn, dialect = db_connection

        # Add feature selection as preprocessing step
        sklearn_pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),  # Standardize features
                ("normalizer", MinMaxScaler()),  # Then scale to
                ("regressor", DecisionTreeRegressor(max_depth=3, random_state=42)),
            ]
        )

        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_preds = sklearn_pipeline.predict(X)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)

        sql_results = execute_sql(sql, conn, dialect, df)
        np.testing.assert_allclose(
            sql_results.values.flatten(),
            sklearn_preds.flatten(),
            rtol=1e-4,
            atol=1e-4,
        )

    # Gradient Boosting Tests
    def test_gradient_boosting_classifier(self, iris_data, db_connection):
        """Test a gradient boosting classifier with categorical preprocessing."""
        df, feature_names = iris_data
        conn, dialect = db_connection

        # Create a deterministic categorical feature based on petal_length
        # This creates predictable categories that can be debugged
        def assign_quality(length):
            if length < 3:
                return "low"
            elif length < 5:
                return "medium"
            else:
                return "high"

        df["quality"] = df["petal_length"].apply(assign_quality)

        # Use ColumnTransformer for mixed preprocessing
        preprocessor = ColumnTransformer(
            transformers=[
                ("num", StandardScaler(), feature_names),
                ("cat", OneHotEncoder(), ["quality"]),
            ]
        )

        sklearn_pipeline = Pipeline(
            [
                ("preprocessor", preprocessor),
                (
                    "classifier",
                    GradientBoostingClassifier(
                        n_estimators=10, max_depth=3, random_state=42
                    ),
                ),
            ]
        )

        # Use all classes, not just binary
        X = df[feature_names + ["quality"]]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_class = sklearn_pipeline.predict(X)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        features["quality"] = types.StringColumnType()
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)
        sql_results = execute_sql(sql, conn, dialect, df)

        np.testing.assert_allclose(
            sql_results["output_label"].to_numpy(), sklearn_class
        )

        sklearn_proba = sklearn_pipeline.predict_proba(X)
        sklearn_proba_df = pd.DataFrame(
            sklearn_proba, columns=sklearn_pipeline.classes_, index=df.index
        )
        for class_label in sklearn_pipeline.classes_:
            np.testing.assert_allclose(
                sql_results[f"output_probability.{class_label}"].to_numpy(),
                sklearn_proba_df[class_label].values.flatten(),
                rtol=1e-4,
                atol=1e-4,
            )

    def test_gradient_boosting_regressor(self, iris_data, db_connection):
        """Test a gradient boosting regressor with standardization."""
        df, feature_names = iris_data
        conn, dialect = db_connection

        # Add StandardScaler as preprocessing
        sklearn_pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),
                (
                    "regressor",
                    GradientBoostingRegressor(
                        n_estimators=10, max_depth=3, random_state=42
                    ),
                ),
            ]
        )

        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_preds = sklearn_pipeline.predict(X)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)
        sql_results = execute_sql(sql, conn, dialect, df)

        np.testing.assert_allclose(
            sql_results["variable"].to_numpy(),
            sklearn_preds.flatten(),
            rtol=1e-4,
            atol=1e-4,
        )

    # Random Forest Tests
    def test_random_forest_classifier(self, iris_data, db_connection):
        """Test a random forest classifier with mixed preprocessing."""
        df, feature_names = iris_data
        conn, dialect = db_connection

        # Create a deterministic categorical feature based on sepal_width
        # This creates predictable regions that can be debugged
        def assign_region(width):
            if width < 3.0:
                return "north"
            elif width < 3.4:
                return "east"
            elif width < 3.8:
                return "south"
            else:
                return "west"

        # Apply to the full dataset (all classes)
        df["region"] = df["sepal_width"].apply(assign_region)

        # Use ColumnTransformer to handle mixed data types
        preprocessor = ColumnTransformer(
            transformers=[
                ("num", StandardScaler(), feature_names),
                ("cat", OneHotEncoder(), ["region"]),
            ]
        )

        sklearn_pipeline = Pipeline(
            [
                ("preprocessor", preprocessor),
                (
                    "classifier",
                    RandomForestClassifier(
                        n_estimators=10, max_depth=3, random_state=42
                    ),
                ),
            ]
        )

        # Use all classes, not just binary
        X = df[feature_names + ["region"]]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_class = sklearn_pipeline.predict(X)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        features["region"] = types.StringColumnType()
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)
        sql_results = execute_sql(sql, conn, dialect, df)

        np.testing.assert_allclose(
            sql_results["output_label"].to_numpy(), sklearn_class
        )

        sklearn_proba = sklearn_pipeline.predict_proba(X)
        sklearn_proba_df = pd.DataFrame(
            sklearn_proba, columns=sklearn_pipeline.classes_, index=df.index
        )
        for class_label in sklearn_pipeline.classes_:
            np.testing.assert_allclose(
                sql_results[f"output_probability.{class_label}"].to_numpy(),
                sklearn_proba_df[class_label].values.flatten(),
                rtol=1e-3,
                atol=1e-6,
            )

    def test_binary_random_forest_classifier(self, iris_data, db_connection):
        """Test a binary random forest classifier with mixed preprocessing."""
        df, feature_names = iris_data
        conn, dialect = db_connection

        # Add categorical feature for more realistic preprocessing
        binary_df = df[df["target"].isin([0, 1])].copy()
        binary_df["region"] = np.random.choice(
            ["north", "south", "east", "west"], size=binary_df.shape[0]
        )

        # Use ColumnTransformer to handle mixed data types
        preprocessor = ColumnTransformer(
            transformers=[
                ("num", StandardScaler(), feature_names),
                ("cat", OneHotEncoder(), ["region"]),
            ]
        )

        sklearn_pipeline = Pipeline(
            [
                ("preprocessor", preprocessor),
                (
                    "classifier",
                    RandomForestClassifier(
                        n_estimators=10, max_depth=3, random_state=42
                    ),
                ),
            ]
        )

        X = binary_df[feature_names + ["region"]]
        y = binary_df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_class = sklearn_pipeline.predict(X)

        features = dict(
            {fname: types.FloatColumnType() for fname in feature_names},
            region=types.StringColumnType(),
        )
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)

        # Test prediction
        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)
        sql_results = execute_sql(sql, conn, dialect, binary_df)
        np.testing.assert_allclose(
            sql_results["output_label"].to_numpy(), sklearn_class
        )

        # Test probabilities with more tolerance for rounding differences
        sklearn_proba = sklearn_pipeline.predict_proba(X)
        sklearn_proba_df = pd.DataFrame(
            sklearn_proba, columns=sklearn_pipeline.classes_
        )
        for class_label in sklearn_pipeline.classes_:
            np.testing.assert_allclose(
                sql_results[f"output_probability.{class_label}"].to_numpy(),
                sklearn_proba_df[class_label].values.flatten(),
                rtol=0.15,  # Increased tolerance for rounding differences in probabilities
                atol=0.15,  # Increased tolerance for rounding differences in probabilities
            )


class TestTreePostTransformations:
    def test_gradient_boosting_binary_classification_post_transform_bug(
        self, db_connection
    ):
        """Test that GBM binary classification applies post-transforms correctly."""
        conn, dialect = db_connection

        # Create mock binary classification data
        X, y = make_classification(n_samples=100, n_features=20, random_state=42)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.2, random_state=42
        )

        sklearn_pipeline = Pipeline(
            [
                (
                    "preprocess",
                    ColumnTransformer(
                        [("scaler", StandardScaler(with_std=False), [0, 1, 2])],
                        remainder="passthrough",
                    ),
                ),
                (
                    "gbm",
                    GradientBoostingClassifier(
                        max_depth=1, n_estimators=1, random_state=42
                    ),
                ),
            ]
        )

        sklearn_pipeline.fit(X_train, y_train)
        sklearn_proba = sklearn_pipeline.predict_proba(X_train)
        sklearn_predictions = sklearn_pipeline.predict(X_train)

        n_cols = len(X_train[0])
        nm_cols = [f"var_{i}" for i in range(n_cols)]
        features = {n: types.DoubleColumnType() for n in nm_cols}

        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)
        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)

        test_data = pd.DataFrame(X_train, columns=nm_cols)
        sql_results = execute_sql(sql, conn, dialect, test_data)

        # Probabilities should sum to 1
        prob_0 = sql_results["output_probability.0"].to_numpy()
        prob_1 = sql_results["output_probability.1"].to_numpy()
        prob_sums = prob_0 + prob_1
        np.testing.assert_allclose(
            prob_sums,
            1.0,
            rtol=1e-6,
            atol=1e-6,
            err_msg="Probabilities don't sum to 1.0 due to incorrect post-transform order in GBM binary classification",
        )

        # Compare individual class probabilities with sklearn
        sklearn_prob_0 = sklearn_proba[:, 0]  # Class 0 probabilities
        sklearn_prob_1 = sklearn_proba[:, 1]  # Class 1 probabilities
        np.testing.assert_allclose(
            prob_1,
            sklearn_prob_1,
            rtol=1e-4,
            atol=1e-4,
            err_msg="Class 1 probabilities don't match sklearn due to incorrect post-transform",
        )
        np.testing.assert_allclose(
            prob_0,
            sklearn_prob_0,
            rtol=1e-4,
            atol=1e-4,
            err_msg="Class 0 probabilities don't match sklearn due to incorrect post-transform order",
        )

        # Predictions should match sklearn
        sql_predictions = sql_results["output_label"].to_numpy()
        np.testing.assert_array_equal(
            sql_predictions,
            sklearn_predictions,
            err_msg="Predictions don't match sklearn due to incorrect probability calculations",
        )

    def test_random_forest_binary_classification_post_transform_check(
        self, db_connection
    ):
        """Test RandomForest binary classification works correctly because it uses post_transform="NONE"."""
        conn, dialect = db_connection

        # Create mock binary classification data
        X, y = make_classification(n_samples=100, n_features=20, random_state=42)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.2, random_state=42
        )

        sklearn_pipeline = Pipeline(
            [
                (
                    "preprocess",
                    ColumnTransformer(
                        [("scaler", StandardScaler(with_std=False), [0, 1, 2])],
                        remainder="passthrough",
                    ),
                ),
                (
                    "rf",
                    RandomForestClassifier(
                        n_estimators=10, max_depth=3, random_state=42
                    ),
                ),
            ]
        )

        sklearn_pipeline.fit(X_train, y_train)
        sklearn_proba = sklearn_pipeline.predict_proba(X_train)
        sklearn_predictions = sklearn_pipeline.predict(X_train)

        n_cols = len(X_train[0])
        nm_cols = [f"var_{i}" for i in range(n_cols)]
        features = {n: types.DoubleColumnType() for n in nm_cols}

        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)
        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)

        test_data = pd.DataFrame(X_train, columns=nm_cols)
        sql_results = execute_sql(sql, conn, dialect, test_data)

        # RandomForest should work correctly: probabilities should sum to 1
        prob_0 = sql_results["output_probability.0"].to_numpy()
        prob_1 = sql_results["output_probability.1"].to_numpy()
        prob_sums = prob_0 + prob_1
        np.testing.assert_allclose(
            prob_sums,
            1.0,
            rtol=1e-6,
            atol=1e-6,
            err_msg="RandomForest: Probabilities don't sum to 1.0 - unexpected bug!",
        )

        # RandomForest probabilities should match sklearn exactly
        sklearn_prob_0 = sklearn_proba[:, 0]  # Class 0 probabilities
        sklearn_prob_1 = sklearn_proba[:, 1]  # Class 1 probabilities
        np.testing.assert_allclose(
            prob_0,
            sklearn_prob_0,
            rtol=1e-4,
            atol=1e-4,
            err_msg="RandomForest: Class 0 probabilities don't match sklearn",
        )

        np.testing.assert_allclose(
            prob_1,
            sklearn_prob_1,
            rtol=1e-4,
            atol=1e-4,
            err_msg="RandomForest: Class 1 probabilities don't match sklearn",
        )

        # RandomForest predictions should match sklearn
        sql_predictions = sql_results["output_label"].to_numpy()
        np.testing.assert_array_equal(
            sql_predictions,
            sklearn_predictions,
            err_msg="RandomForest: Predictions don't match sklearn",
        )

    def test_decision_tree_binary_classification_post_transform_check(
        self, db_connection
    ):
        """Test DecisionTree binary classification works correctly because DecisionTree uses post_transform="NONE"."""
        conn, dialect = db_connection

        # Create mock binary classification data
        X, y = make_classification(n_samples=100, n_features=20, random_state=42)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.2, random_state=42
        )

        sklearn_pipeline = Pipeline(
            [
                (
                    "preprocess",
                    ColumnTransformer(
                        [("scaler", StandardScaler(with_std=False), [0, 1, 2])],
                        remainder="passthrough",
                    ),
                ),
                ("dt", DecisionTreeClassifier(max_depth=5, random_state=42)),
            ]
        )

        sklearn_pipeline.fit(X_train, y_train)
        sklearn_proba = sklearn_pipeline.predict_proba(X_train)
        sklearn_predictions = sklearn_pipeline.predict(X_train)

        n_cols = len(X_train[0])
        nm_cols = [f"var_{i}" for i in range(n_cols)]
        features = {n: types.DoubleColumnType() for n in nm_cols}

        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)
        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)

        test_data = pd.DataFrame(X_train, columns=nm_cols)
        sql_results = execute_sql(sql, conn, dialect, test_data)

        # DecisionTree should work correctly: probabilities should sum to 1
        prob_0 = sql_results["output_probability.0"].to_numpy()
        prob_1 = sql_results["output_probability.1"].to_numpy()
        prob_sums = prob_0 + prob_1
        np.testing.assert_allclose(
            prob_sums,
            1.0,
            rtol=1e-6,
            atol=1e-6,
            err_msg="DecisionTree: Probabilities don't sum to 1.0 - unexpected bug!",
        )

        # DecisionTree probabilities should match sklearn exactly
        sklearn_prob_0 = sklearn_proba[:, 0]  # Class 0 probabilities
        sklearn_prob_1 = sklearn_proba[:, 1]  # Class 1 probabilities
        np.testing.assert_allclose(
            prob_0,
            sklearn_prob_0,
            rtol=1e-4,
            atol=1e-4,
            err_msg="DecisionTree: Class 0 probabilities don't match sklearn",
        )

        np.testing.assert_allclose(
            prob_1,
            sklearn_prob_1,
            rtol=1e-4,
            atol=1e-4,
            err_msg="DecisionTree: Class 1 probabilities don't match sklearn",
        )

        # DecisionTree predictions should match sklearn
        sql_predictions = sql_results["output_label"].to_numpy()
        np.testing.assert_array_equal(
            sql_predictions,
            sklearn_predictions,
            err_msg="DecisionTree: Predictions don't match sklearn",
        )

    def test_gradient_boosting_multiclass_classification_post_transform_check(
        self, iris_data, db_connection
    ):
        """Test that GradientBoosting multi-class classification works correctly with SOFTMAX post-transform."""
        conn, dialect = db_connection

        # Use iris dataset for multi-class classification (3 classes)
        df, feature_names = iris_data

        # Create pipeline with GradientBoosting multi-class classifier
        sklearn_pipeline = Pipeline(
            [
                ("scaler", StandardScaler()),
                (
                    "gbm",
                    GradientBoostingClassifier(
                        n_estimators=5, max_depth=3, random_state=42
                    ),
                ),
            ]
        )

        X = df[feature_names]
        y = df["target"]
        sklearn_pipeline.fit(X, y)
        sklearn_proba = sklearn_pipeline.predict_proba(X)
        sklearn_predictions = sklearn_pipeline.predict(X)

        features = {fname: types.FloatColumnType() for fname in feature_names}
        parsed_pipeline = orbital.parse_pipeline(sklearn_pipeline, features=features)
        sql = orbital.export_sql("data", parsed_pipeline, dialect=dialect)

        test_data = df[feature_names]
        sql_results = execute_sql(sql, conn, dialect, test_data)

        # Multi-class GBM with SOFTMAX should work correctly: probabilities should sum to 1
        prob_0 = sql_results["output_probability.0"].to_numpy()
        prob_1 = sql_results["output_probability.1"].to_numpy()
        prob_2 = sql_results["output_probability.2"].to_numpy()
        prob_sums = prob_0 + prob_1 + prob_2
        np.testing.assert_allclose(
            prob_sums,
            1.0,
            rtol=1e-4,
            atol=1e-4,
            err_msg="Multi-class GBM: Probabilities don't sum to 1.0 - SOFTMAX post-transform bug!",
        )

        # Multi-class GBM probabilities should match sklearn exactly
        sklearn_prob_0 = sklearn_proba[:, 0]  # Class 0 probabilities
        sklearn_prob_1 = sklearn_proba[:, 1]  # Class 1 probabilities
        sklearn_prob_2 = sklearn_proba[:, 2]  # Class 2 probabilities
        np.testing.assert_allclose(
            prob_0,
            sklearn_prob_0,
            rtol=1e-4,
            atol=1e-4,
            err_msg="Multi-class GBM: Class 0 probabilities don't match sklearn",
        )
        np.testing.assert_allclose(
            prob_1,
            sklearn_prob_1,
            rtol=1e-4,
            atol=1e-4,
            err_msg="Multi-class GBM: Class 1 probabilities don't match sklearn",
        )
        np.testing.assert_allclose(
            prob_2,
            sklearn_prob_2,
            rtol=1e-4,
            atol=1e-4,
            err_msg="Multi-class GBM: Class 2 probabilities don't match sklearn",
        )

        # Multi-class GBM predictions should match sklearn
        sql_predictions = sql_results["output_label"].to_numpy()
        np.testing.assert_array_equal(
            sql_predictions,
            sklearn_predictions,
            err_msg="Multi-class GBM: Predictions don't match sklearn",
        )



"""Base class for the translators of each pipeline step."""

import abc
import typing

import ibis
import onnx

from .._utils import onnx as onnx_utils
from .optimizer import Optimizer
from .options import TranslationOptions
from .variables import GraphVariables, VariablesGroup


class Translator(abc.ABC):
    """Base class for all translators.

    This class is responsible for translating pipeline steps into Ibis expressions.
    """

    def __init__(
        self,
        table: ibis.Table,
        node: onnx.NodeProto,
        variables: GraphVariables,
        optimizer: Optimizer,
        options: TranslationOptions,
    ) -> None:
        """
        :param table: The table the generated query should target.
        :param node: The pipeline node to be translated.
        :param variables: The variables used during the translation process.
        :param optimizer: The optimizer used for the translation.
        """
        self._table = table
        self._variables = variables
        self._node = node
        self._optimizer = optimizer
        self._options = options
        self._inputs = node.input
        self._outputs = node.output
        self._attributes = {
            attr.name: onnx_utils.get_attr_value(attr) for attr in node.attribute
        }

    @abc.abstractmethod
    def process(self) -> None:
        """Performs the translation and set the output variable."""
        pass

    @property
    def operation(self) -> str:
        """What is the operation being translated"""
        return self._node.op_type

    @property
    def inputs(self) -> list[str]:
        """The input variables for this node"""
        return [str(i) for i in self._inputs]

    @property
    def outputs(self) -> list[str]:
        """The expected output variables the node should emit"""
        return [str(o) for o in self._outputs]

    @property
    def mutated_table(self) -> ibis.Table:
        """The table as it is being mutated by the translator.

        This is required for the translator to be able too set
        temporary variables that are not part of the final output.

        For example when an expression is used many times,
        the translator can create a temporary column in the
        SQL query to avoid recomputing the same expression.
        That leads to new columns being added to the table.
        """
        return self._table

    def set_output(
        self,
        value: typing.Union[
            ibis.Deferred, ibis.Expr, VariablesGroup, onnx_utils.VariableTypes
        ],
        index: int = 0,
    ) -> None:
        """Set the output variable for the translator.

        This is only allowed if the translator has a single output.
        Otherwise the node is expected to explicitly set every variable.
        """
        if not isinstance(value, (ibis.Expr, VariablesGroup)):
            value = ibis.literal(value)
        self._variables[self.outputs[index]] = value

    def preserve(self, *variables) -> list[ibis.Expr]:
        """Preserve the given variables in the table.

        This causes the variables to be projected in the table,
        so that future expressions can use them instead of
        repeating the expression.
        """
        for v in variables:
            if v.get_name() in self._table.columns:
                raise ValueError(
                    f"Preserve variable already exists in the table: {v.get_name()}"
                )

        mutate_args = {v.get_name(): v for v in variables}
        self._table = self._table.mutate(**mutate_args)

        # TODO: Should probably update self._variables too
        # in case the same variable is used in multiple places
        # but this is not a common case, and it's complex because
        # we don't know the variable name (!= column_name)
        # so we'll leave it for now.
        return [self._table[cname] for cname in mutate_args]

    def variable_unique_short_alias(self, prefix: typing.Optional[str] = None) -> str:
        """Generate a unique short name for a variable.

        This is generally used to generate names for temporary variables
        that are used in the translation process.

        The names are as short as possible to minimize the
        SQL query length.
        """
        shortname = self._variables.generate_unique_shortname()
        if prefix:
            shortname = f"{prefix}_{shortname}"
        return shortname



"""Translation options used to customize the ONNX→SQL conversion."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(slots=True)
class TranslationOptions:
    """Configuration knobs that affect translation behaviour."""

    allow_text_tensors: bool = False



""""""

import typing

import ibis.expr.types

from ..translator import Translator
from ..variables import ValueVariablesGroup, VariablesGroup


class ArrayFeatureExtractorTranslator(Translator):
    """Processes an ArrayFeatureExtractor node and updates the variables with the output expression.

    ArrayFeatureExtractor can be considered the opposit of [orbital.translation.steps.concat.ConcatTranslator][], as
    in most cases it will be used to pick one or more features out of a group of column
    previously concatenated, or to pick a specific feature out of the result of an ArgMax operation.

    The provided indices always refer to the **last** axis of the input tensor.
    If the input is a 2D tensor, the last axis is the column axis. So an index
    of ``0`` would mean the first column. If the input is a 1D tensor instead the
    last axis is the row axis. So an index of ``0`` would mean the first row.

    This could be confusing because axis are inverted between tensors and orbital column groups.
    In the case of Tensors, index=0 means row=0, while instead in orbital
    column groups (by virtue of being a group of columns), index=0 means
    the first column.

    We have to consider that the indices we receive, in case of column groups,
    are actually column indices, not row indices as in case of a tensor,
    the last index would be the column index. In case of single columns,
    instead the index is the index of a row like it would be with a 1D tensor.
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx_aionnxml_ArrayFeatureExtractor.html

        data = self._variables.consume(self.inputs[0])
        indices = self._variables.consume(self.inputs[1])

        if isinstance(data, VariablesGroup):
            # We are selecting a set of columns out of a column group

            # This expects that dictionaries are sorted by insertion order
            # AND that all values of the dictionary are columns.
            data_keys: list[str] = list(data.keys())
            data_values: list[ibis.Expr] = list(data.values())
            single_item: bool = False

            if isinstance(indices, (int, float)):
                indices = [int(indices)]
                single_item = True

            if not isinstance(indices, (list, tuple)):
                raise ValueError(
                    "ArrayFeatureExtractor expects a list of indices as input."
                )

            indices = typing.cast(list[int], indices)
            if len(indices) > len(data_keys):
                raise ValueError(
                    "Indices requested are more than the available numer of columns."
                )

            # Pick only the columns that are in the list of indicies.
            if single_item:
                result = data_values[indices[0]]
            else:
                result = ValueVariablesGroup(
                    {data_keys[i]: data_values[i] for i in indices}
                )
        elif isinstance(data, (tuple, list)):
            # We are selecting values out of a list of values
            # This is usually used to select "classes" out of a list of
            # possible values based on the variables that represents those classes.
            if not isinstance(indices, ibis.expr.types.Column):
                raise ValueError(
                    "ArrayFeatureExtractor expects a column as indices when picking from a group of values."
                )

            result = ibis.cases(
                *((indices == i, col) for i, col in enumerate(data)),
                else_=ibis.null(),
            )
        else:
            raise NotImplementedError(
                "ArrayFeatureExtractor only supports column groups or lists of constants as input."
            )

        self.set_output(result)



import typing

import onnx
import onnx.helper

ListVariableTypes = typing.Union[list[int], list[float], list[str]]
VariableTypes = typing.Union[float, int, str, ListVariableTypes]


def get_initializer_data(var: onnx.TensorProto) -> VariableTypes:
    """Given a constant initializer, return its value"""
    attr_name = onnx.helper.tensor_dtype_to_field(var.data_type)
    values = list(getattr(var, attr_name))
    dimensions = getattr(var, "dims", None)

    if not dimensions and len(values) == 1:
        # If there are no dimensions, it's a scalar
        # and we should return the single value
        return values[0]
    return values


def get_attr_value(attr: onnx.AttributeProto) -> VariableTypes:
    """Given an attribute, return its value"""
    # TODO: Check if it can be replaced with onnx.helper.get_attribute_value
    if attr.type == attr.INTS:
        return list(attr.ints)
    elif attr.type == attr.FLOATS:
        return list(attr.floats)
    elif attr.type == attr.STRINGS:
        return [s.decode("utf-8") if isinstance(s, bytes) else s for s in attr.strings]
    elif attr.type == attr.INT:
        return attr.i
    elif attr.type == attr.FLOAT:
        return attr.f
    elif attr.type == attr.STRING:
        return attr.s.decode("utf-8") if isinstance(attr.s, bytes) else attr.s
    else:
        raise ValueError(f"Unsupported attribute type: {attr.type}")



"""Implementation of the LinearClassifier operator."""

import typing

import ibis

from ..transformations import apply_post_transform
from ..translator import Translator
from ..variables import NumericVariablesGroup, ValueVariablesGroup, VariablesGroup


class LinearClassifierTranslator(Translator):
    """Processes a LinearClassifier node and updates variables with the classification results.

    The LinearClassifier operator computes classification outputs as:
    Scores = X * coefficients + intercepts

    For more complex pipelines the LinearClassifier operator is not always used,
    usually a combination of Mul and Add operations is used.
    """

    def process(self) -> None:
        """Performs the translation and sets the output variables Y (predictions) and Z (scores)."""
        # https://onnx.ai/onnx/operators/onnx_aionnxml_LinearClassifier.html
        coefficients = typing.cast(list[float], self._attributes["coefficients"])
        intercepts = typing.cast(list[float], self._attributes.get("intercepts", []))
        multi_class = typing.cast(int, self._attributes.get("multi_class", 0))
        post_transform = typing.cast(
            str, self._attributes.get("post_transform", "NONE")
        )

        if multi_class != 0:
            raise NotImplementedError("Multi-class classification is not implemented.")

        classlabels: typing.Union[list[str], list[int], None] = typing.cast(
            typing.Optional[list[int]], self._attributes.get("classlabels_ints")
        ) or typing.cast(
            typing.Optional[list[str]], self._attributes.get("classlabels_strings")
        )

        if classlabels is None:
            raise ValueError(
                "LinearClassifier: classlabels_ints or classlabels_strings must be defined."
            )

        if len(self._inputs) != 1:
            raise ValueError("LinearClassifier node must have exactly 1 input.")

        input_operand = self._variables.consume(self._inputs[0])

        # Standardize input_operand to a columns group,
        # so that we can reuse a single implementation.
        if not isinstance(input_operand, VariablesGroup):
            input_operand = ValueVariablesGroup({"feature": input_operand})

        num_features = len(input_operand)
        num_classes = len(classlabels)

        if len(coefficients) != num_classes * num_features:
            raise ValueError(
                "Coefficients length must equal number of classes × number of input fields."
            )

        fieldsgroup = NumericVariablesGroup(input_operand)
        fields = list(fieldsgroup.values())
        scores = []

        for class_idx in range(num_classes):
            start = class_idx * num_features
            end = start + num_features
            coef_slice = coefficients[start:end]
            intercept = intercepts[class_idx] if intercepts else 0.0

            score = ibis.literal(intercept)
            for val, coef in zip(fields, coef_slice):
                score += val * coef

            score = apply_post_transform(score, post_transform)
            scores.append(self._optimizer.fold_operation(score))

        scores_struct = ValueVariablesGroup(
            {str(label): score for label, score in zip(classlabels, scores)}
        )

        max_score = ibis.greatest(*scores_struct.values())
        predictions = ibis.cases(
            *((score == max_score, label) for label, score in scores_struct.items()),
            else_=ibis.null(),
        )

        self.set_output(predictions, index=0)
        self.set_output(scores_struct, index=1)



"""Translator for Concat and FeatureVectorizer operations."""

import typing

import ibis

from ..translator import Translator
from ..variables import ValueVariablesGroup, VariablesGroup


class ConcatTranslator(Translator):
    """Concatenate multiple columns into a single group of columns.

    In tensor terms, this is meant to create a new tensor by concatenating
    the inputs along a given axis. In most cases, this is used to
    concatenate multiple features into a single one, thus its purpose
    is usually to create a column group from separate columns.

    This means that the most common use case is axis=1,
    which means concatenating over the columns (by virtue of
    column/rows in tensors being flipped over column groups),
    and thus only axis=1 case is supported.
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx__Concat.html

        # Currently only support concatenating over columns,
        # we can't concatenate rows.
        if self._attributes["axis"] not in (1, -1):
            # -1 means last axis, which for 2D entities is equal axis=1
            raise NotImplementedError(
                "Concat currently only supports concatenating over columns (axis=1 or -1)."
            )
        self.set_output(self._concatenate_columns(self))

    @classmethod
    def _concatenate_columns(cls, translator: Translator) -> VariablesGroup:
        """Implement actual operation of concatenating columns.

        This is used by both Concat and FeatureVectorizer translators,
        as they both need to concatenate columns.
        """
        result = ValueVariablesGroup()

        for col in translator.inputs:
            feature = translator._variables.consume(col)
            if isinstance(feature, dict):
                # When the feature is a dictionary,  it means that it was previously
                # concatenated with other features. In pure ONNX terms it would be
                # a tensor, so when we concatenate it we should just merge all the values
                # like we would do when concatenating two tensors.
                for key in feature:
                    varname = col + "." + key
                    result[varname] = feature[key]
            elif isinstance(feature, ibis.Expr):
                result[col] = feature
            else:
                raise ValueError(
                    f"Concat: expected a column group or a single column. Got {type(feature)}"
                )

        return result


class FeatureVectorizerTranslator(Translator):
    """Concatenate multiple columns into a single group of columns.

    This is similar to Concat, but it is a simplified version
    that always only acts on columns, and does not support
    concatenating over rows. While Concat can in theory
    support rows concatenation, even though orbital doesn't implement it.
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx_aionnxml_FeatureVectorizer.html

        # We can support this by doing the same as Concat,
        # in most cases it's sufficient
        ninputdimensions = typing.cast(list[int], self._attributes["inputdimensions"])

        if len(ninputdimensions) != len(self._inputs):
            raise ValueError(
                "Number of input dimensions should be equal to number of inputs."
            )

        # Validate that dimensions are actually correct,
        # as inputdimensions is meant to provide the number of columns of each variable
        for input_idx, colname in enumerate(self.inputs):
            dimensions = ninputdimensions[input_idx]
            feature = self._variables.peek_variable(colname)
            if isinstance(feature, dict):
                if len(feature) != dimensions:
                    raise ValueError(
                        f"Number of columns in input {colname} should be equal to the number of dimensions, got {len(feature)} != {dimensions}"
                    )
            else:
                if dimensions != 1:
                    raise ValueError(
                        f"When merging over individual columns, the dimension should be 1, got {dimensions} for {colname}"
                    )

        self.set_output(ConcatTranslator._concatenate_columns(self))



"""Translators for Cast and CastLike operations"""

import typing

import ibis
import onnx

from ..translator import Translator
from ..variables import ValueVariablesGroup, VariablesGroup

ONNX_TYPES_TO_IBIS: dict[int, ibis.expr.datatypes.DataType] = {
    onnx.TensorProto.FLOAT: ibis.expr.datatypes.float32,  # 1: FLOAT
    onnx.TensorProto.DOUBLE: ibis.expr.datatypes.float64,  # 11: DOUBLE
    onnx.TensorProto.STRING: ibis.expr.datatypes.string,  # 8: STRING
    onnx.TensorProto.INT64: ibis.expr.datatypes.int64,  # 7: INT64
    onnx.TensorProto.BOOL: ibis.expr.datatypes.boolean,  # 9: BOOL
}


class CastTranslator(Translator):
    """Processes a Cast node and updates the variables with the output expression.

    Cast operation is used to convert a variable from one type to another one
    provided by the attribute `to`.
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx__Cast.html
        expr = self._variables.consume(self.inputs[0])
        to_type: int = typing.cast(int, self._attributes["to"])
        if to_type not in ONNX_TYPES_TO_IBIS:
            raise NotImplementedError(f"Cast: type {to_type} not supported")

        target_type = ONNX_TYPES_TO_IBIS[to_type]

        def _is_numeric_or_bool(value: ibis.Value) -> bool:
            vtype = value.type()
            return (
                hasattr(vtype, "is_numeric")
                and vtype.is_numeric()
                or hasattr(vtype, "is_boolean")
                and vtype.is_boolean()
            )

        if (
            self._options.allow_text_tensors is False
            and target_type == ibis.expr.datatypes.string
        ):
            # When sklearn2onnx needs to concatenate features into a single tensor
            # it homogenizes their dtype (e.g. casts numeric target-encoder output to
            # string so it can sit alongside passthrough string columns). Besides being
            # redundant for SQL consumers, promoting one column to text forces the whole
            # encoded block to become text as well, so we drop it unless the caller
            # explicitly opts in.
            if isinstance(expr, VariablesGroup):
                if all(_is_numeric_or_bool(expr.as_value(k)) for k in expr):
                    self.set_output(expr)
                    return
            elif isinstance(expr, ibis.Value) and _is_numeric_or_bool(expr):
                self.set_output(expr)
                return

        if isinstance(expr, VariablesGroup):
            casted = ValueVariablesGroup(
                {
                    k: self._optimizer.fold_cast(expr.as_value(k).cast(target_type))
                    for k in expr
                }
            )
            self.set_output(casted)
        elif isinstance(expr, ibis.Value):
            self.set_output(self._optimizer.fold_cast(expr.cast(target_type)))
        else:
            raise ValueError(
                f"Cast: expected a column group or a single column. Got {type(expr)}"
            )


class CastLikeTranslator(Translator):
    """Processes a CastLike node and updates the variables with the output expression.

    CastLike operation is used to convert a variable from one type to
    the same type of another variable, thus uniforming the two
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx__CastLike.html

        # Cast a variable to have the same type of another variable.
        # For the moment provide a very minimal implementation,
        # in most cases this is used to cast concatenated features to the same type
        # of another feature.
        expr = self._variables.consume(self.inputs[0])
        like_expr = self._variables.consume(self.inputs[1])

        # Assert that the first input is a dict (multiple concatenated columns).
        if not isinstance(expr, VariablesGroup):
            # TODO: Support single variables as well.
            #       This should be fairly straightforward to implement,
            #       but there hasn't been the need for it yet.
            raise NotImplementedError(
                "CastLike currently only supports casting a group of columns."
            )

        # Assert that the second input is a single expression.
        if isinstance(like_expr, VariablesGroup):
            raise NotImplementedError(
                "CastLike currently only supports casting to a single column type, not a group."
            )

        if not isinstance(like_expr, ibis.Value):
            raise ValueError(
                f"CastLike: expected a single column. Got {type(like_expr)}"
            )

        # Get the target type from the second input.
        target_type: ibis.DataType = like_expr.type()

        # Now cast each field in the dictionary to the target type.
        casted = ValueVariablesGroup(
            {
                key: self._optimizer.fold_cast(expr.as_value(key).cast(target_type))
                for key in expr
            }
        )
        self.set_output(casted)



"""Defines the translation step for the Gather operation."""

from orbital.translation.variables import VariablesGroup

from ..translator import Translator


class GatherTranslator(Translator):
    """Processes a Gather node and updates the variables with the output expression.

    The gather operations is meant to pick a specific value out of a column or
    column group.

    The first operand can be a column group or a single column,
    while the second operand must be a constant value.

    When the first operand is a column, the second operand must be 0 as
    there is only one column.

    The operation could in theory be used to pick a specific row of columns
    by setting axis=0, but this is not supported in the current implementation.
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx__Gather.html

        axis = self._attributes.get("axis", 0)
        if axis != 1:
            raise NotImplementedError(
                f"Gather: axis {axis} not supported, only selecting columns (axis=1) is supported"
            )

        expr = self._variables.consume(self.inputs[0])
        idx = self._variables.get_initializer_value(self.inputs[1])
        if not isinstance(idx, (tuple, list)) or len(idx) != 1:
            raise NotImplementedError(
                "Gather second operand must a list of one element"
            )

        idx = idx[0]  # TODO: Support gathering multiple columns
        if not isinstance(idx, int):
            raise ValueError("Gather: index must be an integer constant")

        if isinstance(expr, VariablesGroup):
            keys = list(expr.keys())
            if idx < 0 or idx >= len(keys):
                raise IndexError("Gather: index out of bounds")
            self.set_output(expr[keys[idx]])
        else:
            # Assume that if it's a single column by virtue of the fact that we only
            # support axis=1, then the index must be 0.
            if idx != 0:
                raise NotImplementedError(
                    f"Gather: index {idx} not supported for single columns"
                )
            self.set_output(expr)



"""Translate scikit-learn models to an intermediate represetation.

The IR is what will be processed to generate the SQL queries.
"""

import logging
import pickle
from typing import Any, cast

import onnx as _onnx
import skl2onnx as _skl2o
import skl2onnx.convert
import sklearn.pipeline

from ._utils import repr_pipeline
from .types import ColumnType, FeaturesTypes

log = logging.getLogger(__name__)


class ParsedPipeline:
    """An intermediate representation of a scikit-learn pipeline.

    This object can be converted to a SQL query and run on a database.
    It can also be saved and loaded back in binary format for the sake
    of model distribution. Even though distributing the SQL query
    is usually more convenient.
    """

    _model: _onnx.ModelProto  # type: ignore[assignment]
    features: FeaturesTypes  # type: ignore[assignment]

    def __init__(self) -> None:
        """[orbital.ast.ParsedPipeline][] objects can only be created by the [orbital.ast.parse_pipeline][] function."""

        raise NotImplementedError(
            "parse_pipeline must be used to create a ParsedPipeline object."
        )

    @classmethod
    def _from_onnx_model(
        cls,
        model: _onnx.ModelProto,
        features: FeaturesTypes,
    ) -> "ParsedPipeline":
        """Create a [orbital.ast.ParsedPipeline][] from an ONNX model.

        This is considered an internal implementation detail
        as ONNX should never be exposed to the user.

        Returns a new [orbital.ast.ParsedPipeline][] instance.

        :param model: The ONNX model proto to wrap
        :param features: Dictionary mapping feature names to their [orbital.types.ColumnType][] objects
        """
        self = super().__new__(cls)
        self._model = model
        self.features = self._validate_features(features)
        return self

    @classmethod
    def _validate_features(cls, features: FeaturesTypes) -> FeaturesTypes:
        """Validate the features of the pipeline.

        This checks that the features provided are compatible
        with what a SQL query can handle.

        Returns the validated features dictionary.

        :param features: Dictionary mapping feature names to their [orbital.types.ColumnType][] objects
        """
        for name in features:
            if "." in name:
                raise ValueError(
                    f"Feature names cannot contain '.' characters: {name}, replace with '_'"
                )

        for ftype in features.values():
            if not isinstance(ftype, ColumnType):
                raise TypeError(f"Feature types must be ColumnType objects: {ftype}")

        return features

    def dump(self, filename: str) -> None:
        """Dump the parsed pipeline to a file.

        :param filename: Path to the file where the pipeline will be saved
        """
        # While the ONNX model is in protobuf format, and thus
        # it would make sense to use protobuf to serialize the
        # headers too. Using pickle avoids the need to define
        # a new protobuf schema for the headers and compile .proto files.
        header = {"version": 1, "features": self.features}
        header_data = pickle.dumps(header)
        header_len = len(header_data).to_bytes(4, "big")
        with open(filename, "wb") as f:
            f.write(header_len)
            f.write(header_data)
            f.write(self._model.SerializeToString())

    @classmethod
    def load(cls, filename: str) -> "ParsedPipeline":
        """Load a parsed pipeline from a file.

        Returns a [orbital.ast.ParsedPipeline][] object loaded from the specified file.

        :param filename: Path to the file containing the saved pipeline
        """
        with open(filename, "rb") as f:
            header_len = int.from_bytes(f.read(4), "big")
            header_data = f.read(header_len)
            header = pickle.loads(header_data)
            if header["version"] != 1:
                # Currently there is only version 1
                raise UnsupportedFormatVersion("Unsupported format version.")
            model = _onnx.load_model(f)
        return cls._from_onnx_model(model, header["features"])

    def __str__(self) -> str:
        """Generate a string representation of the pipeline."""
        return str(repr_pipeline.ParsedPipelineStr(self))


def parse_pipeline(
    pipeline: sklearn.pipeline.Pipeline, features: FeaturesTypes
) -> ParsedPipeline:
    """Parse a scikit-learn pipeline into an intermediate representation.

    Returns a [orbital.ast.ParsedPipeline][] object that can be converted to SQL queries.

    :param pipeline: The fitted scikit-learn pipeline to parse
    :param features: Mapping of column names to their [orbital.types.ColumnType][] objects from the [orbital.types][] module

    ``features`` should be a mapping of column names that are the inputs of the
    pipeline to their types from the [orbital.types][] module:

    ```
        {
            "column_name": types.DoubleColumnType(),
            "another_column": types.Int64ColumnType()
        }
    ```
    """
    non_passthrough_features = {
        fname: ftype for fname, ftype in features.items() if not ftype.is_passthrough
    }

    if not non_passthrough_features:
        raise ValueError(
            "All provided features are passthrough. "
            "The pipeline would not do anything useful."
        )

    # Check if pipeline starts with a model (which expects concatenated input)
    concatenated_inputs = EnsureConcatenatedInputs(non_passthrough_features)
    pipeline_requires_input_vector = concatenated_inputs.pipeline_requires_input_vector(
        pipeline
    )

    if pipeline_requires_input_vector:
        # Models expect a single feature vector "input", so we need to adapt the user
        # features to a single concatenated input tensor.
        # Later, we'll inject a concat operation to ensure the SQL query does work
        # with individual columns.
        initial_types = concatenated_inputs.concatenate_inputs()
    else:
        initial_types = [
            (fname, ftype._to_onnxtype())
            for fname, ftype in non_passthrough_features.items()
        ]

    onnx_model = cast(
        _onnx.ModelProto,
        _skl2o.to_onnx(pipeline, initial_types=initial_types),  # type: ignore[arg-type]
    )

    if pipeline_requires_input_vector:
        # Inject concat operation to create the "input" tensor when necessary.
        onnx_model = concatenated_inputs.inject_concat_step(onnx_model)

    return ParsedPipeline._from_onnx_model(onnx_model, features)


class EnsureConcatenatedInputs:
    """Handle ONNX input tensor requirements for scikit-learn pipelines.

    ONNX models require a single "input" tensor for models (as opposed to transformers).
    When a pipeline contains only a model without preprocessing steps, sklearn2onnx
    doesn't always automatically add a Concat operation.

    This class provides the necessary logic to:

    1. Detect when a pipeline starts with a model that expects concatenated input
    2. Create proper initial_types for sklearn2onnx with a single concatenated tensor
    3. Inject a Concat operation into the ONNX graph for SQL compatibility

    This bridges the gap between SQL (individual columns) and ONNX models
    (concatenated input tensors).
    """

    def __init__(self, features: FeaturesTypes) -> None:
        """Initialize with the features dictionary.

        :param features: Dictionary mapping feature names to their [orbital.types.ColumnType][] objects
        """
        self.features = features

    def pipeline_requires_input_vector(
        self, pipeline: sklearn.pipeline.Pipeline
    ) -> bool:
        """Determine if pipeline requires concatenated inputs by testing operator compatibility.

        This method directly tests whether the first operator in the pipeline can handle
        individual feature inputs by calling `infer_types`. If it fails, the operator
        requires concatenated inputs.

        Returns True if the pipeline requires concatenated inputs, False otherwise.

        :param pipeline: The scikit-learn pipeline to analyze
        """
        individual_types = [
            (fname, ftype._to_onnxtype()) for fname, ftype in self.features.items()
        ]

        topology = skl2onnx.convert.parse_sklearn_model(
            pipeline, initial_types=individual_types
        )

        if len(self.features) <= 1:
            # The user provided only one feature, no need for concatenation
            return False

        # Get the first operator in the topology
        first_operator = next(topology.unordered_operator_iterator(), None)
        if not first_operator:
            return False

        # Test if the operator can handle the individual inputs we provided
        try:
            first_operator.infer_types()
            # If infer_types() succeeds, the operator accepts the inputs the user provided
            return False
        except RuntimeError as err:
            if "at most 1 input" in str(err):
                # If infer_types() fails with "at most 1 input", the operator needs concatenated inputs
                # This is the best we can do as SKL2ONNX doesn't tell us how many inputs it expects.
                # And the `check_input_and_output_numbers` function always throws a RuntimeError
                return True
            return False

    def concatenate_inputs(self) -> list[tuple[str, Any]]:
        """Create initial_types for skl2onnx when pipeline starts with a model.

        Models expect a single concatenated input tensor, so we create initial_types
        with a single "input" tensor containing all features concatenated together.
        All features must be of the same ONNX type for this to work.

        Returns a list with single tuple: `[("input", onnx_type([None, num_features]))]`.
        """
        # All features must be of the same type for model input
        feature_onnx_types = {
            type(ftype._to_onnxtype()) for ftype in self.features.values()
        }

        if len(feature_onnx_types) != 1:
            # Mixed types not allowed for model input
            type_names = [t.__name__ for t in feature_onnx_types]
            raise ValueError(
                f"All features must be of the same type when pipeline starts with a model. "
                f"Found mixed types: {', '.join(sorted(type_names))}. "
                f"Please ensure all features use the same ColumnType."
            )

        # All features have the same type, use it for concatenated input
        uniform_type = next(iter(feature_onnx_types))
        return [("input", uniform_type([None, len(self.features)]))]

    def inject_concat_step(self, onnx_model: _onnx.ModelProto) -> _onnx.ModelProto:
        """Inject a Concat operation for pipelines starting with models to enable SQL generation.

        Pipelines starting with models create a single "input" tensor, but SQL generation expects
        individual feature columns. This function modifies the ONNX graph to:

        1. Replace the single "input" with individual feature inputs
        2. Add a Concat operation to combine them back into "input"

        This bridges the gap between SQL (individual columns) and models (concatenated input).

        Returns the modified ONNX model with injected Concat operation.

        :param onnx_model: The ONNX model to modify
        """
        graph = onnx_model.graph

        # Verify this is a pipeline with "input" tensor
        input_names = [inp.name for inp in graph.input]
        if "input" not in input_names:
            # Not a model pipeline, return unchanged
            return onnx_model

        feature_names = list(self.features.keys())

        # Create new individual feature inputs
        new_inputs = []

        for fname in feature_names:
            # Create new input tensor for each feature with shape [None, 1]
            ftype = self.features[fname]._to_onnxtype().to_onnx_type()
            new_inputs.append(
                _onnx.helper.make_tensor_value_info(
                    fname,
                    ftype.tensor_type.elem_type,
                    [None, 1],
                )
            )

        # Create Concat node to combine individual features into "input"
        concat_node = _onnx.helper.make_node(
            "Concat",
            inputs=feature_names,  # Individual feature columns
            outputs=["input"],  # Combined input expected by the model
            axis=1,  # Concatenate along feature axis (columns)
            name="orbital_concat",
        )

        # Modify the graph
        # 1. Replace graph inputs with individual features
        del graph.input[:]
        graph.input.extend(new_inputs)

        # 2. Insert concat node at the beginning of the graph
        graph.node.insert(0, concat_node)

        return onnx_model


class UnsupportedFormatVersion(Exception):
    """Format of loaded pipeline is not supported.

    This usually happens when trying to load a newer
    format version with an older version of the framework.
    """

    pass



import numpy as np
import pandas as pd
import pytest
from google.protobuf.json_format import MessageToDict
from sklearn.datasets import make_classification
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.impute import SimpleImputer
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from orbital import ast, types


class TestPipelineParsing:
    DF_DATA = {
        "feature1": [1, 2, 3, np.nan, 5],
        "feature2": [np.nan, 1, 0, 3, 1],
        "feature3": [1.1, 2.1, 3.1, 4.1, np.nan],
    }
    DATA_TYPES = {
        "feature1": types.DoubleColumnType(),
        "feature2": types.DoubleColumnType(),
        "feature3": types.DoubleColumnType(),
    }

    def test_need_to_parse(self):
        with pytest.raises(NotImplementedError):
            ast.ParsedPipeline()

    def test_parse_pipeline(self):
        df = pd.DataFrame(self.DF_DATA)

        pipeline = Pipeline(
            steps=[
                ("imputer", SimpleImputer(strategy="mean")),
                ("scaler", StandardScaler()),
            ]
        )
        pipeline.fit(df)

        parsed = ast.parse_pipeline(pipeline, self.DATA_TYPES)
        assert parsed.features == self.DATA_TYPES
        assert parsed._model is not None

        model_graph = MessageToDict(parsed._model.graph)
        assert {i["name"] for i in model_graph["input"]} == {
            "feature1",
            "feature2",
            "feature3",
        }
        assert {n["name"] for n in model_graph["node"]} == {
            "Di_Div",
            "FeatureVectorizer",
            "Imputer",
            "N1",
            "Su_Sub",
        }

    def test_dump_load_pipeline(self, tmp_path):
        df = pd.DataFrame(self.DF_DATA)

        pipeline = Pipeline(
            steps=[
                ("imputer", SimpleImputer(strategy="mean")),
                ("scaler", StandardScaler()),
            ]
        )
        pipeline.fit(df)

        parsed = ast.parse_pipeline(pipeline, self.DATA_TYPES)
        filename = tmp_path / "test_dump_load_pipeline.dump"
        parsed.dump(filename)

        loaded = ast.ParsedPipeline.load(filename)
        assert loaded.features == parsed.features
        assert loaded._model is not None
        assert loaded._model.SerializeToString() == parsed._model.SerializeToString()

    def test_load_incompatible_version(self, tmp_path):
        import pickle

        header = {"version": 2, "features": {}}
        header_data = pickle.dumps(header)
        header_len = len(header_data).to_bytes(4, "big")

        filename = tmp_path / "test_load_incompatible_version.dump"
        with open(filename, "wb") as f:
            f.write(header_len)
            f.write(header_data)

        with pytest.raises(ast.UnsupportedFormatVersion):
            ast.ParsedPipeline.load(filename)

    def test_parse_pipeline_no_preprocessing(self):
        """Test parsing a pipeline with no preprocessing steps.

        This tests the fix for the issue where sklearn2onnx fails when there are
        no preprocessing steps and multiple input features.
        """
        # Create some test data
        X, y = make_classification(n_features=5, random_state=42)
        X_train, _, y_train, _ = train_test_split(X, y, test_size=0.2, random_state=42)

        # Create a pipeline with only a model (no preprocessing)
        pipeline = Pipeline(
            [
                (
                    "model",
                    GradientBoostingClassifier(
                        n_estimators=3, max_depth=2, random_state=42
                    ),
                )
            ]
        )
        pipeline.fit(X_train, y_train)

        # Define features for all columns
        features = {f"feature_{i}": types.DoubleColumnType() for i in range(X.shape[1])}

        # This should not raise an exception (previously it would fail)
        parsed = ast.parse_pipeline(pipeline, features)

        assert parsed.features == features
        assert parsed._model is not None

        # The model should have been converted successfully
        model_graph = MessageToDict(parsed._model.graph)

        # With the fix, we should have individual feature inputs for SQL compatibility
        assert len(model_graph["input"]) == 5  # One input per feature
        input_names = {inp["name"] for inp in model_graph["input"]}
        expected_names = {f"feature_{i}" for i in range(5)}
        assert input_names == expected_names

        # Should have a Concat node to combine features and the tree ensemble
        node_types = {n["opType"] for n in model_graph["node"]}
        assert "Concat" in node_types  # Injected for SQL compatibility
        assert "TreeEnsembleClassifier" in node_types



"""Defines the translation step for the ArgMax operation."""

import ibis

from ..translator import Translator


class ArgMaxTranslator(Translator):
    """Processes an ArgMax node and updates the variables with the output expression.

    Given the node to translate, the variables and constants available for
    the translation context, generates a query expression that processes
    the input variables and produces a new output variable that computes
    based on the ArgMax operation.

    The ArgMax implementation is currently limited to emitting a variable
    that represents the index of the column with the maximum value in a group
    of columns. It is not possible to compute the max of a set of rows
    thus axis must be 1 and keepdims must be 1.

    As it computes the maximum out of a set of columns, argmax expects
    a columns group as its input.

    The limitation is due to the fact that we can't mix variables with
    a different amount of rows and MAX(col) would end up producing a single row.
    This is usually ok, as ArgMax is primarily used to pick the values
    with the maximum value out of the features analyzed by the model,
    and thus it is only required to produce a new value for each entry
    on which to perform a prediction/classification (row).
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx__ArgMax.html
        data = self._variables.consume(self.inputs[0])
        axis = self._attributes.get("axis", 1)
        keepdims = self._attributes.get("keepdims", 1)
        select_last_index = self._attributes.get("select_last_index", 0)

        if not isinstance(data, dict):
            # if it's a single column, we can't really do much with it
            # as there aren't other comlumns to compare with.
            raise NotImplementedError(
                "ArgMaxTranslator can only be applied to a group of columns"
            )

        if axis != 1:
            # For axis=0 we would want to return the index of the row
            # with the maximum value, but we don't have a row identifier
            raise NotImplementedError("ArgMaxTranslator only supports axis=1")
        if keepdims != 1:
            raise NotImplementedError(
                "ArgMaxTranslator only supports retaining original dimensions"
            )

        keys = list(data.keys())
        if len(keys) == 0:
            raise ValueError("ArgMaxTranslator requires at least one column")

        if len(keys) == 1:
            self.set_output(data[keys[0]])
            return

        # Generate a CASE THEN ELSE expression to find
        # which out of all the columns has the maximum value.
        cases: list[tuple[ibis.Expr, int]] = []
        for idx, key in enumerate(keys):
            cond = None
            # Compare the current column with all other columns
            for j, other in enumerate(keys):
                if j == idx:
                    # Do not compare to yourself.
                    continue
                # When select_last_index is True
                # We use '>', otherwise '>=' so that we can pick the first occurrence.
                cmp_expr = (
                    data[key] > data[other]
                    if select_last_index
                    else data[key] >= data[other]
                )
                cond = cmp_expr if cond is None else cond & cmp_expr
            cases.append((cond, idx))
        argmax_expr = ibis.cases(*cases, else_=0)

        self.set_output(argmax_expr)



"""Test the transformations module functionality."""

import pytest
import ibis
import math

from orbital.translation.transformations import apply_post_transform
from orbital.translation.variables import ValueVariablesGroup, NumericVariablesGroup


def test_apply_post_transform_single_value():
    """Test applying transformations to a single numeric value."""
    # Create a simple numeric value
    value = ibis.literal(2.0)

    # Test NONE transformation (identity)
    result = apply_post_transform(value, "NONE")
    assert isinstance(result, ibis.expr.types.NumericValue)

    # Test LOGISTIC transformation
    result = apply_post_transform(value, "LOGISTIC")
    assert isinstance(result, ibis.expr.types.NumericValue)


def test_apply_post_transform_variables_group():
    """Test applying transformations to a VariablesGroup."""
    # Create a VariablesGroup with numeric values
    group_data = {
        "class_0": ibis.literal(1.0),
        "class_1": ibis.literal(-0.5),
        "class_2": ibis.literal(2.0),
    }
    variables_group = ValueVariablesGroup(group_data)

    # Test NONE transformation
    result = apply_post_transform(variables_group, "NONE")
    assert isinstance(result, NumericVariablesGroup)
    assert len(result) == 3
    assert "class_0" in result
    assert "class_1" in result
    assert "class_2" in result

    # Test LOGISTIC transformation
    result = apply_post_transform(variables_group, "LOGISTIC")
    assert isinstance(result, NumericVariablesGroup)
    assert len(result) == 3


def test_apply_post_transform_invalid_transform():
    """Test that invalid transformations raise NotImplementedError."""
    value = ibis.literal(1.0)

    with pytest.raises(
        NotImplementedError, match="Post transform 'INVALID' is not implemented"
    ):
        apply_post_transform(value, "INVALID")


def test_apply_post_transform_preserves_group_type():
    """Test that the function preserves the exact type of VariablesGroup."""
    # Create a NumericVariablesGroup
    group_data = {
        "value1": ibis.literal(1.0),
        "value2": ibis.literal(2.0),
    }
    numeric_group = NumericVariablesGroup(group_data)

    # Apply transformation
    result = apply_post_transform(numeric_group, "LOGISTIC")

    # Check that the result is still a NumericVariablesGroup
    assert isinstance(result, NumericVariablesGroup)
    assert len(result) == 2


def test_logistic_transform_computes_correct_values():
    """Test that LOGISTIC transformation computes the expected sigmoid values."""
    # Test with a simple DuckDB backend to compute actual values
    backend = ibis.duckdb.connect()

    # Test LOGISTIC transformation on single value
    test_value = 2.0
    input_expr = ibis.literal(test_value)
    logistic_result = apply_post_transform(input_expr, "LOGISTIC")

    # Compute the actual result
    computed_value = backend.execute(logistic_result)
    expected_value = 1 / (1 + math.exp(-test_value))  # sigmoid function

    assert abs(computed_value - expected_value) < 1e-10, (
        f"LOGISTIC: expected {expected_value}, got {computed_value}"
    )

    # Test LOGISTIC transformation on NumericVariablesGroup
    test_values = {"class_0": 1.0, "class_1": -0.5, "class_2": 0.0}
    group_data = {name: ibis.literal(val) for name, val in test_values.items()}
    numeric_group = NumericVariablesGroup(group_data)

    logistic_group = apply_post_transform(numeric_group, "LOGISTIC")

    # Check each value in the group
    for name, expected_input in test_values.items():
        computed_output = backend.execute(logistic_group[name])
        expected_output = 1 / (1 + math.exp(-expected_input))

        assert abs(computed_output - expected_output) < 1e-10, (
            f"LOGISTIC group {name}: expected {expected_output}, got {computed_output}"
        )


def test_none_transform_computes_correct_values():
    """Test that NONE transformation returns identity values (unchanged)."""
    # Test with a simple DuckDB backend to compute actual values
    backend = ibis.duckdb.connect()

    # Test NONE transformation on single value
    test_value = 2.0
    input_expr = ibis.literal(test_value)
    none_result = apply_post_transform(input_expr, "NONE")
    computed_none = backend.execute(none_result)

    assert computed_none == test_value, (
        f"NONE: expected {test_value}, got {computed_none}"
    )

    # Test NONE transformation on NumericVariablesGroup
    test_values = {"class_0": 1.0, "class_1": -0.5, "class_2": 0.0}
    group_data = {name: ibis.literal(val) for name, val in test_values.items()}
    numeric_group = NumericVariablesGroup(group_data)

    none_group = apply_post_transform(numeric_group, "NONE")

    # Check each value in the group remains unchanged
    for name, expected_value in test_values.items():
        computed_value = backend.execute(none_group[name])

        assert computed_value == expected_value, (
            f"NONE group {name}: expected {expected_value}, got {computed_value}"
        )


def test_softmax_transform_computes_correct_values():
    """Test that SOFTMAX transformation computes the expected normalized values."""
    # Test with a simple DuckDB backend to compute actual values
    backend = ibis.duckdb.connect()

    # Test SOFTMAX transformation on single value (should return 1.0)
    test_value = 2.0
    input_expr = ibis.literal(test_value)
    softmax_result = apply_post_transform(input_expr, "SOFTMAX")

    computed_value = backend.execute(softmax_result)
    assert computed_value == 1.0, (
        f"SOFTMAX single value: expected 1.0, got {computed_value}"
    )

    # Test SOFTMAX transformation on NumericVariablesGroup
    test_values = {"class_0": 1.0, "class_1": 2.0, "class_2": 3.0}
    group_data = {name: ibis.literal(val) for name, val in test_values.items()}
    numeric_group = NumericVariablesGroup(group_data)

    softmax_group = apply_post_transform(numeric_group, "SOFTMAX")

    # Check that the result is a NumericVariablesGroup
    assert isinstance(softmax_group, NumericVariablesGroup)
    assert len(softmax_group) == 3

    # Compute actual softmax values manually for verification
    import math

    values = list(test_values.values())
    max_val = max(values)
    exp_values = [math.exp(v - max_val) for v in values]
    sum_exp = sum(exp_values)
    expected_softmax = [exp_val / sum_exp for exp_val in exp_values]

    # Check each computed value
    computed_values = []
    for name in ["class_0", "class_1", "class_2"]:
        computed_val = backend.execute(softmax_group[name])
        computed_values.append(computed_val)

    # Verify the computed softmax values match expected
    for i, (name, expected) in enumerate(
        zip(["class_0", "class_1", "class_2"], expected_softmax)
    ):
        assert abs(computed_values[i] - expected) < 1e-10, (
            f"SOFTMAX group {name}: expected {expected}, got {computed_values[i]}"
        )

    # Verify that the softmax values sum to 1.0
    total_sum = sum(computed_values)
    assert abs(total_sum - 1.0) < 1e-10, (
        f"SOFTMAX values should sum to 1.0, got {total_sum}"
    )



"""Implementation of the OneHotEncoder operator."""

import typing

import ibis

from ..translator import Translator
from ..variables import ValueVariablesGroup


class OneHotEncoderTranslator(Translator):
    """Processes a MatMul node and updates the variables with the output expression.

    Given a categorical variable, this class creates a new group of columns,
    with one column for each category. The values of the column are 1.0
    if the original column value is equal to the category, and 0.0 otherwise.

    It supports only strings for categories and emits floats as column
    values.
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx_aionnxml_OneHotEncoder.html
        cats = typing.cast(list[str], self._attributes.get("cats_strings"))
        if not isinstance(cats, list):
            # We currently only support string values for categories
            raise ValueError("OneHotEncoder: attribute cats_strings not found")

        input_expr = self._variables.consume(self.inputs[0])
        if not isinstance(input_expr, ibis.Value):
            raise ValueError("OneHotEncoder: input expression not found")

        casted_variables = [
            ibis.ifelse(input_expr == cat, 1, 0)
            .cast("float64")
            .name(self.variable_unique_short_alias("oh"))
            for cat in cats
        ]

        # OneHot encoded features are usually consumed multiple times
        # by subsequent operations, so preserving them makes sense.
        casted_variables = self.preserve(*casted_variables)
        self.set_output(
            ValueVariablesGroup(
                {cat: casted_variables[i] for i, cat in enumerate(cats)}
            )
        )



"""Prase tree definitions and return a graph of nodes."""

import typing

import ibis

from ...translator import Translator
from ...variables import VariablesGroup


class BranchConditionCreator:
    """Factory for branch conditions that reuses preserved feature columns."""

    def __init__(
        self,
        translator: Translator,
        input_expr: typing.Union[ibis.Expr, VariablesGroup],
    ) -> None:
        """Set up reusable branch conditions for tree traversal.

        :param translator: The [orbital.translation.translator.Translator][] instance
            that preserves intermediate expressions and holds ONNX node metadata.
        :param input_expr: The expression or
            [orbital.translation.variables.VariablesGroup][] providing the features
            consumed by the tree.
        """
        self._translator = translator
        self._ordered_features = self._prepare_features(input_expr)
        self._conditions = self._prepare_conditions()

    def _prepare_features(
        self, input_expr: typing.Union[ibis.Expr, VariablesGroup]
    ) -> list[ibis.Expr]:
        if isinstance(input_expr, VariablesGroup):
            features: list[ibis.Expr] = input_expr.values_value()
        else:
            features = typing.cast(list[ibis.Expr], [input_expr])

        renamed_features = [
            feature.name(self._translator.variable_unique_short_alias("cnf"))
            for feature in features
        ]
        # `node["feature_id"]` is a 0-based index into this preserved list, so we
        # return it directly instead of building a mapping keyed by feature id.
        return self._translator.preserve(*renamed_features)

    def _condition_expression(
        self,
        feature_expr: ibis.Expr,
        threshold_expr: ibis.Expr,
        mode: str,
    ) -> ibis.Expr:
        if mode == "BRANCH_LEQ":
            return feature_expr <= threshold_expr
        if mode == "BRANCH_LT":
            return feature_expr < threshold_expr
        if mode == "BRANCH_GTE":
            return feature_expr >= threshold_expr
        if mode == "BRANCH_GT":
            return feature_expr > threshold_expr
        if mode == "BRANCH_EQ":
            return feature_expr == threshold_expr
        if mode == "BRANCH_NEQ":
            return feature_expr != threshold_expr

        raise NotImplementedError(f"Unsupported node mode: {mode}")

    def _prepare_conditions(self) -> dict[tuple[int, str, float], ibis.Expr]:
        node_modes = typing.cast(list[str], self._translator._attributes["nodes_modes"])
        node_featureids = typing.cast(
            list[int], self._translator._attributes["nodes_featureids"]
        )
        node_thresholds = typing.cast(
            list[float], self._translator._attributes["nodes_values"]
        )

        keys_to_preserve: dict[tuple[int, str, float], ibis.Expr] = {}

        for feature_id, mode, threshold in zip(
            node_featureids, node_modes, node_thresholds
        ):
            if mode == "LEAF":
                continue
            key = (feature_id, mode, threshold)
            if key in keys_to_preserve:
                continue

            feature_expr = self._ordered_features[feature_id]
            threshold_expr = ibis.literal(threshold)
            raw_condition = self._condition_expression(
                feature_expr, threshold_expr, mode
            )
            named_expression = raw_condition.name(
                self._translator.variable_unique_short_alias("cnc")
            )
            keys_to_preserve[key] = named_expression

        if not keys_to_preserve:
            return {}

        preserved_exprs = self._translator.preserve(*keys_to_preserve.values())
        return dict(zip(keys_to_preserve.keys(), preserved_exprs))

    def create_condition(self, node: dict) -> ibis.Expr:
        """Return a preserved comparison expression for the given branch node."""
        key = (node["feature_id"], node["mode"], node["treshold"])
        return self._conditions[key]


def build_tree(translator: Translator) -> dict[int, dict[int, dict]]:
    """Build a tree based on nested dictionaries of nodes."""
    nodes_treeids = typing.cast(list[int], translator._attributes["nodes_treeids"])
    nodes_nodeids = typing.cast(list[int], translator._attributes["nodes_nodeids"])
    nodes_modes = typing.cast(list[str], translator._attributes["nodes_modes"])
    nodes_truenodeids = typing.cast(
        list[int], translator._attributes["nodes_truenodeids"]
    )
    nodes_falsenodeids = typing.cast(
        list[int], translator._attributes["nodes_falsenodeids"]
    )
    nodes_thresholds = typing.cast(list[float], translator._attributes["nodes_values"])
    nodes_featureids = typing.cast(
        list[int], translator._attributes["nodes_featureids"]
    )
    nodes_missing_value_tracks_true = typing.cast(
        list[int], translator._attributes["nodes_missing_value_tracks_true"]
    )
    node = translator._node

    # Assert a few things to ensure we don't ed up genearting a tree with wrong data
    # All entries related to branches should match in length
    assert (
        len(nodes_treeids)
        == len(nodes_nodeids)
        == len(nodes_modes)
        == len(nodes_truenodeids)
        == len(nodes_falsenodeids)
        == len(nodes_thresholds)
        == len(nodes_featureids)
    )

    # Weight could be a float or a dictionary of class labels weights
    weights: dict = {}
    if node.op_type == "TreeEnsembleClassifier":
        weights = typing.cast(
            dict[tuple[int, int], dict[typing.Union[str, int], float]], weights
        )
        # Weights for classifier, in this case the weights are per-class
        class_nodeids = typing.cast(list[int], translator._attributes["class_nodeids"])
        class_treeids = typing.cast(list[int], translator._attributes["class_treeids"])
        class_weights = typing.cast(
            list[float], translator._attributes["class_weights"]
        )
        weights_classid = typing.cast(list[int], translator._attributes["class_ids"])
        assert (
            len(class_treeids)
            == len(class_nodeids)
            == len(class_weights)
            == len(weights_classid)
        )
        classlabels = typing.cast(
            typing.Optional[list[typing.Union[str, int]]],
            translator._attributes.get("classlabels_strings")
            or translator._attributes.get("classlabels_int64s"),
        )
        if not classlabels:
            raise ValueError("Missing class labels when building tree")

        for tree_id, node_id, weight, weight_classid in zip(
            class_treeids, class_nodeids, class_weights, weights_classid
        ):
            node_weights = typing.cast(
                dict[typing.Union[str, int], float],
                weights.setdefault((tree_id, node_id), {}),
            )
            node_weights[classlabels[weight_classid]] = weight

    elif node.op_type == "TreeEnsembleRegressor":
        # Weights for the regressor, in this case leaf nodes have only 1 weight
        weights = typing.cast(dict[tuple[int, int], float], weights)
        target_weights = typing.cast(
            list[float], translator._attributes["target_weights"]
        )
        target_nodeids = typing.cast(
            list[int], translator._attributes["target_nodeids"]
        )
        target_treeids = typing.cast(
            list[int], translator._attributes["target_treeids"]
        )
        assert len(target_treeids) == len(target_nodeids) == len(target_weights)
        for tree_id, node_id, weight in zip(
            target_treeids, target_nodeids, target_weights
        ):
            weights[(tree_id, node_id)] = weight
    else:
        raise NotImplementedError(f"Unsupported tree node type: {node.op_type}")

    # Create all nodes for the trees
    trees: dict[int, dict[int, dict]] = {}
    for tree_id, node_id, mode, true_id, false_id, threshold, feature_id in zip(
        nodes_treeids,
        nodes_nodeids,
        nodes_modes,
        nodes_truenodeids,
        nodes_falsenodeids,
        nodes_thresholds,
        nodes_featureids,
    ):
        if tree_id not in trees:
            trees[tree_id] = {}

        node_dict = {
            "id": (tree_id, node_id),
            "mode": mode,
            "feature_id": feature_id,
            "missing_tracks_true": bool(
                nodes_missing_value_tracks_true[node_id]
                if nodes_missing_value_tracks_true
                else 0
            ),
        }
        if mode == "LEAF":
            node_dict["weight"] = weights[(tree_id, node_id)]
        else:
            node_dict["treshold"] = threshold

        trees[tree_id][node_id] = node_dict

    # Link nodes creating a tree structure
    for tree_id, node_id, true_id, false_id in zip(
        nodes_treeids,
        nodes_nodeids,
        nodes_truenodeids,
        nodes_falsenodeids,
    ):
        if node_id in trees[tree_id]:
            node_dict = trees[tree_id][node_id]
            if node_dict["mode"] == "LEAF":
                # Leaf nodes have no true or false branches
                # In the end they are leaves so they don't have branches
                continue
            if true_id in trees[tree_id]:
                node_dict["true"] = trees[tree_id][true_id]
            if false_id in trees[tree_id]:
                node_dict["false"] = trees[tree_id][false_id]

    return {tree_id: trees[tree_id][0] for tree_id in trees}



import operator

import ibis
import pytest

from orbital.translation.optimizer import Optimizer


class TestOptimizerFold:
    optimizer = Optimizer()

    def test_fold_sum_only_literals(self):
        result = self.optimizer.fold_contiguous_sum(
            [ibis.literal(1), ibis.literal(2), ibis.literal(3)]
        )
        assert len(result) == 1
        assert result[0].execute() == pytest.approx(6)

    def test_fold_sum_with_column_and_literals(self):
        table = ibis.memtable({"value": [1.0, 2.0, 3.0]})
        column = table["value"]
        result = self.optimizer.fold_contiguous_sum(
            [ibis.literal(1), column, ibis.literal(2)]
        )
        assert len(result) == 2
        assert result[0] is column
        assert result[1].execute() == pytest.approx(3)

    def test_fold_sum_preserves_non_scalar_order(self):
        table = ibis.memtable({"a": [1.0], "b": [2.0]})
        a_col = table["a"]
        b_col = table["b"]
        result = self.optimizer.fold_contiguous_sum([a_col, ibis.literal(5), b_col])
        assert len(result) == 3
        assert result[0] is a_col
        assert result[1] is b_col
        assert result[2].execute() == pytest.approx(5)

    def test_fold_sum_keeps_identity_with_other_terms(self):
        table = ibis.memtable({"value": [1.0, 2.0, 3.0]})
        column = table["value"]
        result = self.optimizer.fold_contiguous_sum([ibis.literal(0), column])
        assert len(result) == 2
        assert result[0] is column
        assert result[1].execute() == pytest.approx(0)

    def test_fold_product_returns_zero_when_zero_present(self):
        table = ibis.memtable({"value": [1.0, 2.0, 3.0]})
        column = table["value"]
        result = self.optimizer.fold_contiguous_product(
            [column, ibis.literal(0), ibis.literal(5)]
        )
        assert len(result) == 2
        assert result[0] is column
        assert result[1].execute() == pytest.approx(0)

    def test_fold_product_with_column_and_literals(self):
        table = ibis.memtable({"value": [1.0, 2.0, 3.0]})
        column = table["value"]
        result = self.optimizer.fold_contiguous_product(
            [ibis.literal(2), column, ibis.literal(3)]
        )
        assert len(result) == 2
        assert result[0] is column
        assert result[1].execute() == pytest.approx(6)

    def test_fold_product_preserves_non_scalar_order(self):
        table = ibis.memtable({"a": [1.0], "b": [2.0]})
        a_col = table["a"]
        b_col = table["b"]
        result = self.optimizer.fold_contiguous_product([a_col, ibis.literal(4), b_col])
        assert len(result) == 3
        assert result[0] is a_col
        assert result[1] is b_col
        assert result[2].execute() == pytest.approx(4)

    def test_fold_product_keeps_identity_with_other_terms(self):
        table = ibis.memtable({"value": [1.0, 2.0, 3.0]})
        column = table["value"]
        result = self.optimizer.fold_contiguous_product([ibis.literal(1), column])
        assert len(result) == 2
        assert result[0] is column
        assert result[1].execute() == pytest.approx(1)

    def test_fold_unsupported_operator_raises(self):
        with pytest.raises(NotImplementedError):
            self.optimizer._fold_associative_op_contiguous(
                [ibis.literal(1), ibis.literal(2)], operator.sub
            )

    def test_fold_cast_merges_nested_casts(self):
        table = ibis.memtable({"a": [1.0]})
        expr = table["a"].cast("float32").cast("string")
        folded = self.optimizer.fold_cast(expr)
        assert isinstance(folded.op(), ibis.expr.operations.Cast)
        assert not isinstance(folded.op().arg, ibis.expr.operations.Cast)
        assert folded.type().is_string()

    def test_fold_cast_literal_int(self):
        expr = ibis.literal(3.7).cast("int64")
        result = self.optimizer.fold_cast(expr)
        assert result.op().value == 3

    def test_fold_cast_literal_float(self):
        expr = ibis.literal(5).cast("float64")
        result = self.optimizer.fold_cast(expr)
        assert result.op().value == 5.0

    def test_fold_cast_literal_string(self):
        expr = ibis.literal(42).cast("string")
        result = self.optimizer.fold_cast(expr)
        assert result.op().value == "42"

    def test_fold_cast_literal_boolean(self):
        expr = ibis.literal(1).cast("boolean")
        result = self.optimizer.fold_cast(expr)
        assert result.op().value is True

    def test_fold_cast_nested_literals(self):
        expr = ibis.literal(7).cast("float64").cast("string")
        result = self.optimizer.fold_cast(expr)
        assert result.op().value == "7"

    def test_fold_zeros_subtract_left(self):
        expr = ibis.literal(0) - ibis.literal(5)
        result = self.optimizer.fold_zeros(expr)
        assert result.op().value == 5

    def test_fold_zeros_subtract_right(self):
        expr = ibis.literal(5) - ibis.literal(0)
        result = self.optimizer.fold_zeros(expr)
        assert result.op().value == 5

    def test_fold_operation_unary_negate(self):
        expr = -ibis.literal(10)
        result = self.optimizer.fold_operation(expr)
        assert result == -10

    def test_fold_operation_unary_not(self):
        expr = ~ibis.literal(True)
        result = self.optimizer.fold_operation(expr)
        assert result is False



"""Implementation of the Identity operator."""

from ..translator import Translator


class IdentityTranslator(Translator):
    """Processes an Identity node and updates the variables with the output expression.

    The identity node is a no-op, it simply passes the input to the output,
    it is meant to copy the input into the output, but as there could be
    multiple references to the same expression, it doesn't actually need
    to perform a copy.
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx__Identity.html

        self.set_output(self._variables.consume(self._inputs[0]))



"""Implementation of the Where operator."""

import itertools

import ibis

from ..translator import Translator
from ..variables import ValueVariablesGroup, VariablesGroup


class WhereTranslator(Translator):
    """Processes a Where node and updates the variables with the output expression.

    The where operation is expected to return ether its first or second input
    depending on a condition variable. When the variable is true, the first
    input is returned, otherwise the second input is returned.

    The condition variable will usually be a column computed through an expression
    that represents a boolean predicate.

    The first and second inputs can be either a single column or a group of columns.
    If any of the two is a group of columns, a new group of column is produced
    as the result.
    If both are single columns, the result is a single column.
    """

    def process(self) -> None:
        """Performs the translation and set the output variable."""
        # https://onnx.ai/onnx/operators/onnx__Where.html
        condition_expr = self._variables.consume(self.inputs[0])
        true_expr = self._variables.consume(self.inputs[1])
        false_expr = self._variables.consume(self.inputs[2])

        if isinstance(condition_expr, VariablesGroup):
            raise NotImplementedError(
                "Where: The condition expression can't be a group of columns. Must be a single column."
            )

        if isinstance(true_expr, VariablesGroup) and isinstance(
            false_expr, VariablesGroup
        ):
            true_values = list(true_expr.values())
            false_values = list(false_expr.values())
            if len(true_values) != len(false_values):
                raise ValueError(
                    "Where: The number of values in the true and false expressions must match."
                )
            result = ValueVariablesGroup()
            for true_val, false_val, idx in zip(
                true_values, false_values, itertools.count()
            ):
                result[f"c{idx}"] = self._optimizer.fold_case(
                    ibis.cases((condition_expr, true_val), else_=false_val)
                )
        elif isinstance(true_expr, VariablesGroup) and not isinstance(
            false_expr, VariablesGroup
        ):
            result = ValueVariablesGroup()
            for idx, true_val in enumerate(true_expr.values()):
                result[f"c{idx}"] = self._optimizer.fold_case(
                    ibis.cases((condition_expr, true_val), else_=false_expr)
                )
        elif not isinstance(true_expr, VariablesGroup) and isinstance(
            false_expr, VariablesGroup
        ):
            result = ValueVariablesGroup()
            for idx, false_val in enumerate(false_expr.values()):
                result[f"c{idx}"] = self._optimizer.fold_case(
                    ibis.cases((condition_expr, true_expr), else_=false_val)
                )
        else:
            result = self._optimizer.fold_case(
                ibis.cases((condition_expr, true_expr), else_=false_expr)
            )

        self.set_output(result)
